import pya
import math
from pya import *
from SiEPIC.utils import get_technology_by_name


def layout_Ring(cell, layer, x, y, r, w, npoints):
    # function to produce the layout of a ring resonator
    # cell: layout cell to place the layout
    # layer: which layer to use
    # x, y: location of the origin
    # r: radius
    # w: waveguide width
    # units in microns

    # example usage.  Places the ring layout in the presently selected cell.
    # cell = Application.instance().main_window().current_view().active_cellview().cell
    # layout_Ring(cell, cell.layout().layer(TECHNOLOGY['Si']), 0, 0, 10, 0.5, 400)

    # fetch the database parameters
    dbu = cell.layout().dbu

    # compute the circle
    pts = []
    da = math.pi * 2 / npoints
    for i in range(0, npoints + 1):
        pts.append(
            Point.from_dpoint(
                DPoint(
                    (x + (r + w / 2) * math.cos(i * da)) / dbu,
                    (y + (r + w / 2) * math.sin(i * da)) / dbu,
                )
            )
        )
    for i in range(npoints, -1, -1):
        pts.append(
            Point.from_dpoint(
                DPoint(
                    (x + (r - w / 2) * math.cos(i * da)) / dbu,
                    (y + (r - w / 2) * math.sin(i * da)) / dbu,
                )
            )
        )

    # create the shape
    cell.shapes(layer).insert(Polygon(pts))

    # end of layout_Ring


class DoubleBus_Ring(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the double bus ring resonator.
    Consists of a ring with 2 straight waveguides.
    """

    def __init__(self):
        # Important: initialize the super class
        super(DoubleBus_Ring, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("s", self.TypeShape, "", default=DPoint(0, 0))
        self.param("r", self.TypeDouble, "Radius", default=10)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("g", self.TypeDouble, "Gap", default=0.2)
        self.param(
            "textpolygon", self.TypeInt, "Draw text polygon label? 0/1", default=1
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "DoubleBus_Ring(R="
            + ("%.3f" % self.r)
            + ",g="
            + ("%g" % (1000 * self.g))
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.silayer
        LayerSiN = ly.layer(LayerSi)
        TextLayerN = ly.layer(self.textl)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.utils import points_per_circle, layout_pgtext

        # Create the ring resonator:
        layout_Ring(
            self.cell,
            LayerSiN,
            self.r + self.w / 2,
            self.r + self.g + self.w,
            self.r,
            self.w,
            points_per_circle(self.r),
        )

        w = int(round(self.w / dbu))
        r = int(round(self.r / dbu))
        g = int(round(self.g / dbu))

        #   pcell = ly.create_cell("DirectionalCoupler_HalfRing_Straight", "SiEPIC", { "r": self.r, "w": self.w, "g": self.g, "silayer": LayerSi, "bustype": 0 } )
        #   print ("Cell: pcell: #%s" % pcell.cell_index())
        #   t = Trans(Trans.R0, 0, 0)
        #   instance = self.cell.insert(CellInstArray(pcell.cell_index(), t))
        #   t = Trans(Trans.R180, 0, 2*r+2*g+2*w)
        #   instance = self.cell.insert(CellInstArray(pcell.cell_index(), t))

        # Create the two waveguides
        wg1 = Box(0, -w / 2, w + 2 * r, w / 2)
        shapes(LayerSiN).insert(wg1)
        y_offset = 2 * r + 2 * g + 2 * w
        wg2 = Box(0, y_offset - w / 2, w + 2 * r, y_offset + w / 2)
        shapes(LayerSiN).insert(wg2)

        from SiEPIC._globals import PIN_LENGTH as pin_length
        # Create the pins, as short paths:

        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, 0, 0)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(w + 2 * r - pin_length / 2, 0),
                Point(w + 2 * r + pin_length / 2, 0),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, w + 2 * r, 0)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [Point(pin_length / 2, y_offset), Point(-pin_length / 2, y_offset)], w
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, 0, y_offset)
        text = Text("pin3", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(w + 2 * r - pin_length / 2, y_offset),
                Point(w + 2 * r + pin_length / 2, y_offset),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        t = Trans(Trans.R0, w + 2 * r, y_offset)
        text = Text("pin4", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer
        dev = Box(0, -w * 3, w + 2 * r, y_offset + w * 3)
        shapes(LayerDevRecN).insert(dev)

        # Add a polygon text description
        if self.textpolygon:
            layout_pgtext(
                self.cell,
                self.textl,
                self.w,
                self.r + self.w,
                "%.3f-%g" % (self.r, self.g),
                1,
            )

        # print("Done drawing the layout for - DoubleBus_Ring: %.3f-%g" % ( self.r, self.g) )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_irph_mrr(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the IRPH MRR.

    Author: Jaspreet Jhoja
    jaspreetj@ece.ubc.ca
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_irph_mrr, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "siriblayer",
            self.TypeLayer,
            "Si rib Layer",
            default=TECHNOLOGY["Si - 90 nm rib"],
        )
        self.param("r", self.TypeDouble, "Radius", default=15)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("g", self.TypeDouble, "Gap", default=0.156)
        self.param(
            "io_wg_type",
            self.TypeBoolean,
            "I/O waveguide type, 0 - strip; 1 - rib",
            default=0,
        )

        self.param("n_w", self.TypeDouble, "N Width", default=31)
        self.param("n_theta_start", self.TypeDouble, "Start Theta", default=-25)
        self.param("n_theta_stop", self.TypeDouble, "Stop Theta", default=115)

        self.param("npp_si", self.TypeDouble, "N++ to Si edge", default=2.0)
        self.param("npp_w", self.TypeDouble, "N++ Donut Width", default=13.5)

        self.param("vc_cw", self.TypeDouble, "VC Center radius", default=5.75)
        self.param("vc_aw", self.TypeDouble, "VC Arc Width", default=5)

        self.param("m_cw", self.TypeDouble, "Metal Center radius", default=10)
        self.param("m_aw", self.TypeDouble, "Metal Arc Width", default=11)
        self.param(
            "overlay_ebl", self.TypeDouble, "Overlay accuracy (EBL) (um)", default=0.05
        )

        self.param("nlayer", self.TypeLayer, "N Layer", default=TECHNOLOGY["Si N"])
        self.param(
            "npplayer", self.TypeLayer, "N++ Layer", default=TECHNOLOGY["Si N++"]
        )
        self.param("vclayer", self.TypeLayer, "VC Layer", default=TECHNOLOGY["VC"])
        self.param(
            "mlayer", self.TypeLayer, "Metal Layer", default=TECHNOLOGY["M2_router"]
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "ebeam_irph_mrr(R="
            + ("%.3f" % self.r)
            + ",g="
            + ("%g" % (1000 * self.g))
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        from SiEPIC.utils import arc_wg_xy, arc_xy
        from SiEPIC._globals import PIN_LENGTH
        from SiEPIC.extend import to_itype

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerSiRibN = ly.layer(self.siriblayer)
        LayerNN = ly.layer(self.nlayer)
        LayerNPPN = ly.layer(self.npplayer)
        LayerVCN = ly.layer(self.vclayer)
        LayerMN = ly.layer(self.mlayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        w = to_itype(self.w, dbu)
        r = to_itype(self.r, dbu)
        g = to_itype(self.g, dbu)

        n_w = to_itype(self.n_w, dbu)
        npp_si = to_itype(self.npp_si, dbu)
        npp_w = to_itype(self.npp_w, dbu)
        theta_start = self.n_theta_start
        theta_stop = self.n_theta_stop

        vc_cw = to_itype(self.vc_cw, dbu)
        vc_aw = to_itype(self.vc_aw, dbu)
        overlay_ebl = to_itype(self.overlay_ebl, dbu)

        m_cw = to_itype(self.m_cw, dbu)
        m_aw = to_itype(self.m_aw, dbu)
        x = 0
        y = 0

        # draw ring
        self.cell.shapes(LayerSiN).insert(arc_wg_xy(x, y, r, w, 0, 360))

        # draw bus waveguides
        # bottom waveguide
        xtop = -2 / 3 * r
        ytop = -1 * (r + g + w / 2)
        xbottom = 4 / 3 * r
        ybottom = ytop - w
        wg1 = Box(xtop, ytop, xbottom, ybottom)
        shapes(LayerSiN).insert(wg1)
        # left waveguide
        wg1 = Box(ytop, xtop, ybottom, xbottom)
        shapes(LayerSiN).insert(wg1)

        # Pins on the bottom waveguide side:
        pin = Path(
            [
                Point(xtop + PIN_LENGTH, ytop - w / 2),
                Point(xtop - PIN_LENGTH, ytop - w / 2),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin1", Trans(Trans.R0, xtop, ytop - w / 2))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(xbottom - PIN_LENGTH, ytop - w / 2),
                Point(xbottom + PIN_LENGTH, ytop - w / 2),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin2", Trans(Trans.R0, xbottom, ytop - w / 2))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Pins on the left waveguide side:
        pin = Path(
            [
                Point(ytop - w / 2, xtop + PIN_LENGTH),
                Point(ytop - w / 2, xtop - PIN_LENGTH),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin3", Trans(Trans.R0, ytop - w / 2, xtop))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(ybottom + w / 2, xbottom - PIN_LENGTH),
                Point(ybottom + w / 2, xbottom + PIN_LENGTH),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin4", Trans(Trans.R0, ybottom + w / 2, xbottom))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # draw N
        arc_pts = arc_xy(x, y, n_w, theta_start, theta_stop)
        arc_pts.append(
            pya.Point.from_dpoint(pya.DPoint(0, 0))
        )  # adding center point for polygon
        shapes(LayerNN).insert(pya.Polygon(arc_pts))

        # draw NPP
        # center
        arc_pts = arc_xy(x, y, r - npp_si - w / 2, 0, 360)
        arc_pts.append(
            pya.Point.from_dpoint(pya.DPoint(0, 0))
        )  # adding center point for polygon
        shapes(LayerNPPN).insert(pya.Polygon(arc_pts))
        # outer donut
        shapes(LayerNPPN).insert(
            arc_wg_xy(
                x, y, r + w / 2 + npp_si + npp_w / 2, npp_w, theta_start, theta_stop
            )
        )

        # draw via
        # center
        arc_pts = arc_xy(x, y, vc_cw, 0, 360)
        arc_pts.append(
            pya.Point.from_dpoint(pya.DPoint(0, 0))
        )  # adding center point for polygon
        shapes(LayerVCN).insert(pya.Polygon(arc_pts))

        # outer donut
        shapes(LayerVCN).insert(
            arc_wg_xy(
                x, y, 4 / dbu + r + npp_w / 2, vc_aw, theta_start + 10, theta_stop - 10
            )
        )

        # draw metal
        # center
        arc_pts = arc_xy(x, y, m_cw, 0, 360)
        arc_pts.append(
            pya.Point.from_dpoint(pya.DPoint(0, 0))
        )  # adding center point for polygon
        shapes(LayerMN).insert(pya.Polygon(arc_pts))
        # outer donut
        shapes(LayerMN).insert(
            arc_wg_xy(x, y, 4 / dbu + r + npp_w / 2, m_aw, theta_start, theta_stop)
        )

        # devrec layer
        xtop = -2 / 3 * r
        ytop = -1 * (r + g + w)
        xbottom = 4 / 3 * r
        ybottom = ytop
        dev_array = [
            Point(ytop + 3.5 * w, xbottom + npp_w),
            Point(ytop + 3.5 * w, xbottom),
            Point(ytop - 3.5 * w, xbottom),
            Point(ytop - 3.5 * w, xtop),
            Point(ytop + 3.5 * w, xtop),
            Point(xtop, ybottom + 3.5 * w),
            Point(xtop, ybottom - 3 * w),
            Point(xbottom, ybottom - 3.5 * w),
            Point(xbottom, ybottom + 3.5 * w),
            Point(xbottom + npp_w, ybottom + 3.5 * w),
            Point(xbottom + npp_w, xbottom + npp_w),
        ]

        shapes(LayerDevRecN).insert(pya.Polygon(dev_array))

        if self.io_wg_type == 0:  # strip converter
            taper_long = to_itype(10, dbu)
            taper_short = to_itype(3, dbu)
            dev_array = [
                Point(ytop + 3.5 * w, xbottom + npp_w),
                Point(ytop + 3.5 * w, xbottom - taper_long),
                Point(ytop + 2.5 * w, xbottom - taper_long),
                Point(ytop + w / 2 - overlay_ebl * 2, xbottom),
                Point(ytop - w / 2 + overlay_ebl * 2, xbottom),
                Point(ytop - 2.5 * w, xbottom - taper_long),
                Point(ytop - 3.5 * w, xbottom - taper_long),
                Point(ytop - 3.5 * w, xtop + taper_short),
                Point(ytop - 1.5 * w, xtop + taper_short),
                Point(ytop - w / 2 + overlay_ebl * 2, xtop),
                Point(ytop + w / 2 - overlay_ebl * 2, xtop),
                Point(ytop + 1.5 * w, xtop + taper_short),
                Point(ytop + 3.5 * w, xtop + taper_short),
                Point((ytop + 3.5 * w + xtop) / 2, (xtop + ybottom + 3.5 * w) / 2),
                Point(xtop + taper_short, ybottom + 3.5 * w),
                Point(xtop + taper_short, ybottom + 1.5 * w),
                Point(xtop, ybottom + w / 2 - overlay_ebl * 2),
                Point(xtop, ybottom - w / 2 + overlay_ebl * 2),
                Point(xtop + taper_short, ybottom - 1.5 * w),
                Point(xtop + taper_short, ybottom - 3.5 * w),
                Point(xbottom - taper_long, ybottom - 3.5 * w),
                Point(xbottom - taper_long, ybottom - 2.5 * w),
                Point(xbottom, ybottom - w / 2 + overlay_ebl * 2),
                Point(xbottom, ybottom + w / 2 - overlay_ebl * 2),
                Point(xbottom - taper_long, ybottom + 2.5 * w),
                Point(xbottom - taper_long, ybottom + 3.5 * w),
                Point(xbottom + npp_w, ybottom + 3.5 * w),
                Point(xbottom + npp_w, xbottom + npp_w),
            ]
        # Si rib layer
        shapes(LayerSiRibN).insert(pya.Polygon(dev_array))

        t = Trans(Trans.R0, 0, 0)
        text = Text("Component=ebeam_irph_mrr", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = self.r * 0.07 / dbu

        # print("Done drawing the layout for - ebeam_IRPH_MRR: %.3f-%g" % ( self.r, self.g))



"""
This file is part of the SiEPIC-EBeam-PDK

Version history:

Lukas Chrostowski 2023/02/19
 - PCell takes an input of the waveguide type from a dropdown list
    Loaded from Waveguides.xml
 - Metal heater
"""

from pya import *
import pya


class wg_heater(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        super(wg_heater, self).__init__()

        from SiEPIC.utils import get_technology_by_name, load_Waveguides_by_Tech

        """
    # https://github.com/KLayout/klayout/issues/879
    tech = self.layout.library().technology
    if not tech:
       tech = 'EBeam'
    self.technology_name = tech
    """
        self.technology_name = "EBeam"
        TECHNOLOGY = get_technology_by_name(self.technology_name)

        # Load all waveguides
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # declare the parameters

        default_wg = "Strip TE 1550 nm, w=500 nm"
        p = self.param(
            "waveguide_type", self.TypeList, "Waveguide Type", default=default_wg
        )
        #    p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default = self.waveguide_types[0]['name'])

        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])
        self.param("length", self.TypeDouble, "Length", default=100)
        self.param(
            "path",
            self.TypeShape,
            "Path",
            default=DPath([DPoint(0, 0), DPoint(100, 0)], 0.5),
            hidden=True,
        )

        self.param("mh_width", self.TypeInt, "Heater width (microns)", default=4)
        self.param(
            "mhlayer", self.TypeLayer, "Heater Layer", default=TECHNOLOGY["M1_heater"]
        )
        self.param(
            "mllayer",
            self.TypeLayer,
            "Metal contact layer",
            default=TECHNOLOGY["M2_router"],
        )
        self.param("metal_width", self.TypeDouble, "Metal contact width", default=10)

        self.param(
            "wg_width",
            self.TypeDouble,
            "Waveguide width (microns)",
            default=0.5,
            readonly=True,
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param(
            "pinrecm",
            self.TypeLayer,
            "PinRecM Layer (metal)",
            default=TECHNOLOGY["PinRecM"],
        )
        self.param("text", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

        self.cellName = "wg_heater"

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "%s_%s" % (self.cellName, self.length)

    def coerce_parameters_impl(self):
        params = [t for t in self.waveguide_types if t["name"] == self.waveguide_type]
        if not params:
            raise Exception("Waveguides.XML not correctly configured")
        self.wg_width = float(params[0]["width"])

    def can_create_from_shape_impl(self):
        return self.shape.is_path()

    def transformation_from_shape_impl(self):
        return Trans(Trans.R0, 0, 0)

    def parameters_from_shape_impl(self):
        self.path = self.shape.path

    def produce_impl(self):
        # https://github.com/KLayout/klayout/issues/879
        # tech = self.layout.library().technology

        # Make sure the technology name is associated with the layout
        #  PCells don't seem to know to whom they belong!
        if self.layout.technology_name == "":
            self.layout.technology_name = self.technology_name

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")
        params = [t for t in self.waveguide_types if t["name"] == self.waveguide_type]
        layer = [wg["layer"] for wg in params[0]["component"]]
        try:
            layer.remove("DevRec")
        except:
            pass
        if not layer:
            raise Exception("Waveguides.XML not correctly configured")

        # fetch design layers
        LayermlN = self.layout.layer(self.mllayer)
        LayermhN = self.layout.layer(self.mhlayer)
        LayerPinRecN = self.layout.layer(self.pinrec)
        LayerDevRecN = self.layout.layer(self.devrec)
        LayerPinRecM = self.layout.layer(self.pinrecm)
        LayerWG = ly.layer(TECHNOLOGY[layer[0]])
        LayerTextN = ly.layer(self.text)

        from SiEPIC.extend import to_itype
        from SiEPIC.utils.layout import make_pin

        dbu = self.layout.dbu
        w = to_itype(self.wg_width, dbu)

        # Draw the waveguide geometry
        path = DPath([DPoint(0, 0), DPoint(self.length, 0)], self.wg_width)
        self.cell.shapes(LayerWG).insert(path.simple_polygon())

        # draw metal heater
        path = DPath([DPoint(0, 0), DPoint(self.length, 0)], self.mh_width)
        self.cell.shapes(LayermhN).insert(path.simple_polygon())

        # metal contact
        self.cell.shapes(LayermlN).insert(
            pya.DPath(
                [
                    pya.DPoint(self.metal_width / 2, -self.mh_width / 2),
                    pya.DPoint(self.metal_width / 2, self.mh_width / 2),
                ],
                self.metal_width,
            )
        )
        self.cell.shapes(LayermlN).insert(
            pya.DPath(
                [
                    pya.DPoint(self.length - self.metal_width / 2, -self.mh_width / 2),
                    pya.DPoint(self.length - self.metal_width / 2, self.mh_width / 2),
                ],
                self.metal_width,
            )
        )

        # metal pins
        mh_width = to_itype(self.mh_width, dbu)
        ml_width = to_itype(self.metal_width, dbu)
        make_pin(
            self.cell,
            "elec1",
            [self.metal_width / 2, self.mh_width / 2],
            ml_width,
            LayerPinRecM,
            90,
        )
        make_pin(
            self.cell,
            "elec2",
            [self.length - self.metal_width / 2, self.mh_width / 2],
            ml_width,
            LayerPinRecM,
            90,
        )

        # Create the pins on the waveguides
        make_pin(self.cell, "opt1", [0, 0], w, LayerPinRecN, 180)
        make_pin(self.cell, "opt2", [to_itype(self.length, dbu), 0], w, LayerPinRecN, 0)

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        cml = "EBeam"
        if cml:
            text = Text("Lumerical_INTERCONNECT_library=Design kits/%s" % cml, t)
            shape = shapes(LayerDevRecN).insert(text)
            shape.text_size = 0.1 / dbu
            t = Trans(Trans.R0, 0, w * 2)
        model = "wg_heater"
        if model:
            text = Text("Component=%s" % model, t)
            shape = shapes(LayerDevRecN).insert(text)
            shape.text_size = 0.1 / dbu
            t = Trans(Trans.R0, 0, -w * 2)
        text = Text("Spice_param:wg_length=%.3fu" % (self.length), t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        h = max(w * 3 * dbu, self.mh_width / 2)
        box1 = Box(0, -to_itype(h, dbu), to_itype(self.length, dbu), to_itype(h, dbu))
        shapes(LayerDevRecN).insert(box1)



"""
Create a layout of a paperclip spiral,
uses: delay line, waveguide loss cutback structure

- this can be executed directly, or loaded by the technology library
- you can resize the PCell by using the Partial select tool "L", and moving
   the GUIDING SHAPE Box
- Waveguide types are loaded from the XML file
- Pitch of the waveguides is determined by the DevRec layer
- When executed directly, the unit test instantiates all the waveguide types

PCell options:
 - waveguide type
 - ports_opposite = True: input on left, output on right
                  = False: both i/o on left
by Lukas Chrostowski, 2023-2024
"""

import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class spiral_paperclip(pya.PCellDeclarationHelper):
    def __init__(self):
        # Important: initialize the super class
        self.cellName = self.__class__.__name__
        super(eval(self.cellName), self).__init__()

        self.technology_name = "EBeam"

        from SiEPIC.utils import load_Waveguides_by_Tech

        self.TECHNOLOGY = get_technology_by_name(self.technology_name)

        # Load all strip waveguides
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # declare the parameters
        p = self.param(
            "waveguide_type",
            self.TypeList,
            "Waveguide Type",
            default=self.waveguide_types[0]["name"],
        )
        for wa in self.waveguide_types:
            p.add_choice(wa["name"], wa["name"])
        self.param(
            "ports_opposite",
            self.TypeBoolean,
            "Waveguide ports on opposite sides",
            default=False,
        )
        self.param("loops", self.TypeInt, "Number of loops", default=2)
        self.minlength = 2 * float(self.waveguide_types[0]["radius"])
        self.param(
            "length",
            self.TypeDouble,
            "Inner length (min 2 x bend radius)",
            default=self.minlength,
        )
        self.param(
            "box",
            self.TypeShape,
            "Box",
            default=pya.DBox(
                -self.minlength, -self.minlength / 2, self.minlength, self.minlength / 2
            ),
        )
        self.param(
            "length1",
            self.TypeDouble,
            "length: for PCell",
            default=self.minlength,
            hidden=True,
        )
        self.param(
            "flatten",
            self.TypeBoolean,
            "Flatten the PCell, for scripting",
            default=False,
        )
        self.param(
            "tot_length",
            self.TypeDouble,
            "Total length estimate (mm)",
            default=0,
            readonly=True,
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "%s_%s_%s_%s" % (
            self.cellName,
            self.loops,
            self.length,
            self.waveguide_type.replace(" ", "_"),
        )

    def get_waveguide_parameters(self):
        """
        Get the Waveguide Type from the PCell, and update waveguide parameters
        """
        self.waveguide_params = [
            t for t in self.waveguide_types if t["name"] == self.waveguide_type
        ]
        if not self.waveguide_params:
            raise Exception("Waveguides.XML not correctly configured")
        self.waveguide_params = self.waveguide_params[0]
        wg_params = self.waveguide_params
        wg_params2 = wg_params

        # check for compound type
        if "compound_waveguide" in wg_params:
            sm_wg = self.waveguide_params["compound_waveguide"]["singlemode"]
            mm_wg = self.waveguide_params["compound_waveguide"]["multimode"]
            wg_params = [t for t in self.waveguide_types if t["name"] == sm_wg][0]
            wg_params2 = [t for t in self.waveguide_types if t["name"] == mm_wg][0]

        # DevRec width
        if "DevRec" not in [wg["layer"] for wg in wg_params2["component"]]:
            from SiEPIC import _globals

            self.devrec = (
                max([float(wg["width"]) for wg in wg_params2["component"]])
                + _globals.WG_DEVREC_SPACE * 2
            )
        else:
            self.devrec = float(
                [f for f in wg_params2["component"] if f["layer"] == "DevRec"][0][
                    "width"
                ]
            )

        # Radius
        self.radius = float(wg_params["radius"])

        # Waveguide pin width
        self.wg_width = float(wg_params["width"])

    def coerce_parameters_impl(self):
        self.get_waveguide_parameters()

        # We employ coerce_parameters_impl to decide whether the handle or the
        # numeric parameters have changed. We update the
        # numerical value or the shape, depending on which on has not changed.
        w = None
        if isinstance(self.box, pya.Box):
            w = self.box.right * self.layout.dbu
            # print('%s Box: width %s ' %(self.cellName,w))
        if isinstance(self.box, pya.DBox):
            w = self.box.right
            # print('%s DBox: width %s ' %(self.cellName,w))
        if w != None and abs(self.length1 - self.length) < 1e-6:
            if w < self.minlength:
                w = self.minlength
            # print('%s update from GUI Box: %s ' %(self.cellName,w))
            self.length = w
            self.length1 = w
        else:
            # print('%s update from PCell parameters: %s ' %(self.cellName,self.length1))
            self.length1 = self.length
        self.box = pya.DBox(
            -self.length, -self.minlength / 2, self.length, self.minlength / 2
        )

        # estimate length
        # estimate total length of the spiral
        length0 = max(self.minlength, self.length)
        devrec = self.devrec
        radius = self.radius

        # SBend offset in the middle of the spiral
        if "sbends" in self.waveguide_params:
            # Use the Bezier S-Bend (more space efficient)
            # if sbends=True in Waveguide.xml
            offset = radius - devrec / 2
            extra = 0
        else:
            # Use 2 x 90 degree bends
            offset = radius
            extra = devrec

        # Ensure the length is sufficient
        self.minlength = 2 * radius

        # min loops
        self.loops = max(self.loops, 1)

        # Calculate the points again to estimate the total length of the spiral using .distance pya method
        points = [
            DPoint(-length0, radius),
            DPoint(0.0, radius),
            DPoint(0.0, -radius),
            DPoint(length0, -radius),
        ]
        for i in range(1, self.loops * 2, 2):
            points.insert(
                0,
                DPoint(
                    -length0 - devrec * (i - 1), radius - radius * 2 - devrec * (i - 1)
                ),
            )
            points.insert(
                0, DPoint(length0 + devrec * i, radius - radius * 2 - devrec * (i - 1))
            )
            points.insert(
                0, DPoint(length0 + devrec * i, -radius + radius * 2 + devrec * i)
            )
            points.insert(
                0,
                DPoint(-length0 - devrec * (i + 1), -radius + radius * 2 + devrec * i),
            )
            points.append(
                DPoint(
                    length0 + devrec * (i - 1), radius * 2 - radius + devrec * (i - 1)
                )
            )
            points.append(
                DPoint(-length0 - devrec * i, radius * 2 - radius + devrec * (i - 1))
            )
            points.append(
                DPoint(-length0 - devrec * i, -radius * 2 + radius - devrec * i)
            )
            points.append(
                DPoint(length0 + devrec * (i + 1), -radius * 2 + radius - devrec * i)
            )
        if not self.ports_opposite:
            points.append(
                DPoint(
                    length0 + devrec * (i + 1), radius * 2 - radius + devrec * (i + 1)
                )
            )
            points.append(
                DPoint(
                    -length0 - devrec * (i + 1), radius * 2 - radius + devrec * (i + 1)
                )
            )
        points.pop(0)
        points.insert(
            0, DPoint(-length0 - devrec * (i + 1), -radius + radius * 2 + devrec * i)
        )

        for j in range(1, len(points)):
            self.tot_length += points[j - 1].distance(points[j])

        # Update the total length parameter
        self.tot_length = self.tot_length / 1000  # Convert to mm

        # self.tot_length = self.length*2 * (2* self.loops+1+ 0 if self.ports_opposite else 1 )/ 1000 # micron to mm.

    def can_create_from_shape_impl(self):
        return self.shape.is_box()

    def transformation_from_shape_impl(self):
        # Implement the "Create PCell from shape" protocol: we use the center of the shape's
        # bounding box to determine the transformation
        return pya.Trans(self.shape.bbox().center())

    def parameters_from_shape_impl(self):
        # self.path = self.shape.path
        # Implement the "Create PCell from shape" protocol: we set r and l from the shape's
        # bounding box width and layer
        self.width = self.shape.bbox().width() * self.layout.dbu
        self.height = self.shape.bbox().height() * self.layout.dbu

    def produce_impl(self):
        self.get_waveguide_parameters()
        devrec = self.devrec
        radius = self.radius

        # SBend offset in the middle of the spiral
        if "sbends" in self.waveguide_params:
            # Use the Bezier S-Bend (more space efficient)
            # if sbends=True in Waveguide.xml
            offset = radius - devrec / 2
            extra = 0
        else:
            # Use 2 x 90 degree bends
            offset = radius
            extra = devrec

        # Ensure the length is sufficient
        self.minlength = 2 * radius
        length0 = max(self.minlength, self.length)

        # min loops
        self.loops = max(self.loops, 1)

        # spiral points
        points = [
            DPoint(-length0, offset),
            DPoint(0.0, offset),
            DPoint(0.0, -offset),
            DPoint(length0, -offset),
        ]
        for i in range(1, self.loops * 2, 2):
            points.insert(
                0,
                DPoint(
                    -length0 - devrec * (i - 1),
                    offset - radius * 2 - devrec * (i - 1) - extra,
                ),
            )
            points.insert(
                0,
                DPoint(
                    length0 + devrec * i, offset - radius * 2 - devrec * (i - 1) - extra
                ),
            )
            points.insert(
                0,
                DPoint(length0 + devrec * i, -offset + radius * 2 + devrec * i + extra),
            )
            points.insert(
                0,
                DPoint(
                    -length0 - devrec * (i + 1),
                    -offset + radius * 2 + devrec * i + extra,
                ),
            )
            points.append(
                DPoint(
                    length0 + devrec * (i - 1),
                    radius * 2 - offset + devrec * (i - 1) + extra,
                )
            )
            points.append(
                DPoint(
                    -length0 - devrec * i,
                    radius * 2 - offset + devrec * (i - 1) + extra,
                )
            )
            points.append(
                DPoint(-length0 - devrec * i, -radius * 2 + offset - devrec * i - extra)
            )
            points.append(
                DPoint(
                    length0 + devrec * (i + 1),
                    -radius * 2 + offset - devrec * i - extra,
                )
            )
        if not self.ports_opposite:
            points.append(
                DPoint(
                    length0 + devrec * (i + 1),
                    radius * 2 - offset + devrec * (i + 1) + extra,
                )
            )
            points.append(
                DPoint(
                    -length0 - devrec * (i + 1),
                    radius * 2 - offset + devrec * (i + 1) + extra,
                )
            )
        points.pop(0)
        points.insert(
            0,
            DPoint(
                -length0 - devrec * (i + 1), -offset + radius * 2 + devrec * i + extra
            ),
        )

        # Create a path and waveguide
        path = DPath(points, 0.5)
        self.layout.technology_name = (
            self.technology_name
        )  # required otherwise "create_cell" doesn't load
        pcell = self.layout.create_cell(
            "Waveguide",
            self.technology_name,
            {"path": path, "waveguide_type": self.waveguide_type},
        )
        t = Trans(Trans.R0, 0, 0)
        self.cell.insert(CellInstArray(pcell.cell_index(), t))

        # If the waveguide is a Compound type:
        # Flatten the PCell and remove the DevRec and PinRec layers
        if "compound_waveguide" in self.waveguide_params and self.flatten:
            self.cell.flatten(True)
            LayerPinRecN = self.layout.layer(self.TECHNOLOGY["PinRec"])
            LayerDevRecN = self.layout.layer(self.TECHNOLOGY["DevRec"])
            self.cell.clear(LayerPinRecN)
            bbox = self.cell.bbox()
            self.cell.clear(LayerDevRecN)
            self.cell.clear(self.layout.layer(self.TECHNOLOGY["Waveguide"]))
            self.cell.shapes(LayerDevRecN).insert(bbox)

            # Create the pins on the input & output waveguides
            from SiEPIC.utils.layout import make_pin

            if self.ports_opposite:
                make_pin(
                    self.cell,
                    "optA",
                    [
                        length0 + devrec * (i + 1),
                        -radius * 2 + offset - devrec * i - extra,
                    ],
                    self.wg_width,
                    LayerPinRecN,
                    0,
                )
            else:
                make_pin(
                    self.cell,
                    "optA",
                    [
                        -length0 - devrec * (i + 1),
                        radius * 2 - offset + devrec * (i + 1) + extra,
                    ],
                    self.wg_width,
                    LayerPinRecN,
                    180,
                )
            make_pin(
                self.cell,
                "optB",
                [
                    -length0 - devrec * (i + 1),
                    -offset + radius * 2 + devrec * i + extra,
                ],
                self.wg_width,
                LayerPinRecN,
                180,
            )


# Save the path, used for loading WAVEGUIDES.XML
import os

filepath = os.path.dirname(os.path.realpath(__file__))


class test_lib(Library):
    def __init__(self):
        tech = "EBeam"
        library = tech + "test_lib"
        self.technology = tech
        self.layout().register_pcell("spiral_paperclip", spiral_paperclip())
        self.register(library)


if __name__ == "__main__":
    print("Test layout for: Spiral Paperclip")

    from SiEPIC.utils.layout import new_layout
    from SiEPIC.scripts import zoom_out

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan
    topcell, ly = new_layout(tech, "test", GUI=True, overwrite=True)
    # floorplan(topcell, 100e3, 100e3)

    # instantiate the cell
    library = tech + "test_lib"

    if 0:
        # default waveguide
        pcell = ly.create_cell("spiral_paperclip", library, {})
        t = Trans(Trans.R0, 0, 0)
        topcell.insert(CellInstArray(pcell.cell_index(), t))

        # multimode
        pcell = ly.create_cell(
            "spiral_paperclip",
            library,
            {
                "waveguide_type": "Si routing TE 1550 nm (compound waveguide)",
                "length": 100,
                "loops": 1,
            },
        )
        t = Trans(Trans.R0, 0, 40000)
        topcell.insert(CellInstArray(pcell.cell_index(), t))

    # Create spirals for all the types of waveguides
    from SiEPIC.utils import load_Waveguides_by_Tech

    waveguide_types = load_Waveguides_by_Tech(tech)
    xmax = 0
    for ports_opposite in [True, False]:
        for flatten in [True, False]:
            y = 0
            x = xmax
            for wg in waveguide_types:
                print(wg)
                pcell = ly.create_cell(
                    "spiral_paperclip",
                    library,
                    {
                        "waveguide_type": wg["name"],
                        "length": 100,
                        "loops": 1,
                        "flatten": flatten,
                        "ports_opposite": ports_opposite,
                    },
                )
                t = Trans(Trans.R0, x, y - pcell.bbox().bottom)
                inst = topcell.insert(CellInstArray(pcell.cell_index(), t))
                y += pcell.bbox().height() + 2000
                xmax = max(xmax, x + inst.bbox().width())

    zoom_out(topcell)



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class ebeam_bragg_apodized(pya.PCellDeclarationHelper):
    """
    Author:   Mustafa Hammood
              Mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_bragg_apodized, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods", self.TypeInt, "Number of grating periods", default=300
        )
        self.param(
            "grating_period", self.TypeDouble, "Grating period (microns)", default=0.317
        )
        self.param(
            "corrugation_width",
            self.TypeDouble,
            "Corrugration width (microns)",
            default=0.05,
        )
        self.param(
            "misalignment",
            self.TypeDouble,
            "Grating misalignment (microns)",
            default=0.0,
        )
        self.param(
            "sinusoidal",
            self.TypeBoolean,
            "Grating Type (Rectangular=False, Sinusoidal=True)",
            default=False,
        )
        self.param("wg_width", self.TypeDouble, "Waveguide width", default=0.5)
        self.param("index", self.TypeDouble, "Gaussian Index", default=3)
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_bragg_apodized_%s-%.3f-%.3f-%.3f" % (
            self.number_of_periods,
            self.grating_period,
            self.corrugation_width,
            self.misalignment,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype

        # Draw the Bragg grating:
        box_width = int(round(self.grating_period / 2 / dbu))
        grating_period = int(round(self.grating_period / dbu))
        w = to_itype(self.wg_width, dbu)
        GaussianIndex = self.index
        half_w = w / 2
        half_corrugation_w = int(round(self.corrugation_width / 2 / dbu))
        misalignment = int(round(self.misalignment / dbu))
        N = self.number_of_periods
        if self.sinusoidal:
            npoints_sin = 40
            for i in range(0, self.number_of_periods):
                x = round((i * self.grating_period) / dbu)
                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = int(round(self.corrugation_width / 2 / dbu)) * profileFunction
                box1 = Box(x, 0, x + box_width, half_w + profile)
                pts1 = [Point(x, 0)]
                pts3 = [Point(x + misalignment, 0)]
                for i1 in range(0, npoints_sin + 1):
                    x1 = i1 * 2 * math.pi / npoints_sin
                    y1 = round(profile * math.sin(x1))
                    x1 = round(x1 / 2 / math.pi * grating_period)
                    #          print("x: %s, y: %s" % (x1,y1))
                    pts1.append(Point(x + x1, half_w + y1))
                    pts3.append(Point(x + misalignment + x1, -half_w - y1))
                pts1.append(Point(x + grating_period, 0))
                pts3.append(Point(x + grating_period + misalignment, 0))
                shapes(LayerSiN).insert(Polygon(pts1))
                shapes(LayerSiN).insert(Polygon(pts3))
            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, 0, length, half_w)
                shapes(LayerSiN).insert(box2)
                # extra piece at the beginning:
                box3 = Box(0, 0, misalignment, -half_w)
                shapes(LayerSiN).insert(box3)

        else:
            for i in range(0, self.number_of_periods):
                x = int(round((i * self.grating_period) / dbu))

                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = int(round(self.corrugation_width / 2 / dbu)) * profileFunction
                box1 = Box(x, 0, x + box_width, half_w + profile)
                box2 = Box(x + box_width, 0, x + grating_period, half_w - profile)
                box3 = Box(
                    x + misalignment, 0, x + box_width + misalignment, -half_w - profile
                )
                box4 = Box(
                    x + box_width + misalignment,
                    0,
                    x + grating_period + misalignment,
                    -half_w + profile,
                )
                shapes(LayerSiN).insert(box1)
                shapes(LayerSiN).insert(box2)
                shapes(LayerSiN).insert(box3)
                shapes(LayerSiN).insert(box4)
            length = x + grating_period + misalignment
            if misalignment > 0:
                # extra piece at the end:
                box2 = Box(x + grating_period, 0, length, half_w)
                shapes(LayerSiN).insert(box2)
                # extra piece at the beginning:
                box3 = Box(0, 0, misalignment, -half_w)
                shapes(LayerSiN).insert(box3)

        # Create the pins on the waveguides, as short paths:
        from SiEPIC._globals import PIN_LENGTH as pin_length

        t = Trans(Trans.R0, 0, 0)
        pin = Path([Point(pin_length / 2, 0), Point(-pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = Trans(Trans.R0, length, 0)
        pin = Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Compact model information
        t = Trans(Trans.R0, 0, 0)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 10, 0)
        text = Text("Component=ebeam_bragg_te1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu
        t = Trans(Trans.R0, length / 9, 0)
        text = Text(
            "Spice_param:number_of_periods=%s grating_period=%.3fu corrugation_width=%.3fu misalignment=%.3fu sinusoidal=%s"
            % (
                self.number_of_periods,
                self.grating_period,
                self.corrugation_width,
                self.misalignment,
                int(self.sinusoidal),
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        t = Trans(Trans.R0, 0, 0)
        path = Path([Point(0, 0), Point(length, 0)], 3 * w)
        shapes(LayerDevRecN).insert(path.simple_polygon())



"""
KLayout-SiEPIC library for photonic crystals, UBC and SFU


*******
PCells:
*******

1) swg_fc
 - sub-wavelength grating (SWG) fibre coupler (FC)


NOTE: after changing the code, the macro needs to be rerun to install the new
implementation. The macro is also set to "auto run" to install the PCell
when KLayout is run.

Version history:

2017/07/07 Timothy Richards (Simon Fraser University, BC, Canada) and Adam DeAbreu (Simon Fraser University, BC, Canada)
 - swg_fc PCell

2017/07/07 Lukas Chrostowski
 - library definition and github repo

2017/07/09 Jaspreet Jhoja
  - Added Cavity Hole Pcell

2017/07/09 Jingda Wu
 - Added 2D H0 Photonic crystal cavity with single bus waveguide and pins

2017/07/10 Megan Nantel
- Added waveguide with impedance matching tapers for transition from external waveguide to Photonic crystal W1 waveguide

2017/07/10 Jingda Wu
- Improved generation efficiency by using single hole as a cell

2017/07/12 Megan Nantel
- Added the H0c test structure that includes grating couplers, waveguides, and H0c

2017/07/12 Jingda Wu
- Added L3 cavity with double bus waveguide and pins
- Added a drop bus to H0 cavity(coupling between waveguide?)
- Simplified code for generation

2017/07/12 Lukas Chrostowski
 - SWGFC litho test structure

2017/07/13 Megan Nantel
- grating coupler to grating coupler reference test structure
- photonic crystal with only W1 waveguide
- photonic crystal W1 reference test structure

2017/07/16 Jingda Wu
- Adaptive cavity generation under difference waveguide location
- Able to choose the number of waveguides per PhC cavity
- Added etch layer (12,0) on PhC slabs
- Added H0 cavity with oxide buffer, reduced the vertices (32->20) for holes due to much smaller hole radius
- Deleteted cavity hole class
- Added hexagon half cell and hexagon with hole half cell for PhC generation, in case needed
- Added H0 cavity generated with hexagon cells
- Added PhC test pattern

2017/08/19 Jingda Wu
- Added suspension anchor areas for the cavities with undercut

2018/02/14 Lukas Chrostowski
- Upgrade to KLayout 0.25 and SiEPIC-Tools v0.3.x, updating layers to SiEPIC-EBeam v0.3.0+

2021/04/01 lukas
- moved phc and swg cells to beta library


"""

# Import KLayout Python API methods:
# Box, Point, Polygon, Text, Trans, LayerInfo, etc
from pya import *
import pya
import math

from SiEPIC.utils import get_technology_by_name
from SiEPIC.utils import points_per_circle  # ,layout


# -------------------------------------------------------------------------------------------------------------------------------------------------- #
# -------------------------------------------------------------------------------------------------------------------------------------------------- #


class PhC_test(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(PhC_test, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param("n", self.TypeInt, "Number of holes in x and y direction", default=5)
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        self.param("n_sweep", self.TypeInt, "Different sizes of holes", default=13)
        self.param("n_vertices", self.TypeInt, "Vertices of a hole", default=32)
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param(
            "etch", self.TypeLayer, "oxide etch layer", default=pya.LayerInfo(12, 0)
        )

    # def display_text_impl(self):
    # Provide a descriptive text for the cell
    # return "PhC resolution test_a%s-r%.3f-n%.3f" % \
    # (self.a, self.r, self.n)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)
        LayerEtch = ly.layer(self.etch)
        TextLayerN = ly.layer(self.textl)

        # Fetch all the parameters:
        a = self.a / dbu
        r = self.r / dbu
        n_vertices = self.n_vertices
        n = int(math.ceil(self.n / 2))
        # print(n)
        n_sweep = self.n_sweep
        n_x = n
        n_y = n

        # Define Si slab and hole region for future subtraction
        Si_slab = pya.Region()
        hole = pya.Region()
        ruler = pya.Region()
        # hole_r = [r+50,r}
        """
      # translate to array (to pv)
      pv = []
      for p in pcell_decl.get_parameters():
        if p.name in param:
          pv.append(param[p.name])
        else:
          pv.append(p.default)

      pcell_var = self.layout.add_pcell_variant(lib, pcell_decl.id(), pv)
      t_text = pya.Trans(x_offset-2*a_k, -y_offset-a_k*0.5)
      self.cell.insert(pya.CellInstArray(pcell_var, t_text))      
      
      for m in range(0,28):        
        ruler.insert(pya.Box(-x_width+x_offset_2+x_spacing*m, -y_height+y_offset, x_width+x_offset_2+x_spacing*m, y_height+y_offset))
        if m > 23:
          None
        else:
          ruler.insert(pya.Box(-y_height+x_offset_3, -x_width-y_offset_2+x_spacing*m, y_height+x_offset_3, x_width-y_offset_2+x_spacing*m))
        
      for j in range(-n_y,n_y+1):
        if j%2 == 0:
          for i in range(-n_x,n_x+1):
            if i!=0:
              hole_x = abs(i)/i*(abs(i)-0.5)*a_k+x_offset
              hole_y = j*a_k*math.sqrt(3)/2
              hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
              hole_t = hole_poly.transformed(hole_trans)
              hole.insert(hole_t)
            #print(hole_t) 
        elif j%2 == 1:
          for i in range(-n_x,n_x+1): 
            hole_x = i*a_k+x_offset
            hole_y = j*a_k*math.sqrt(3)/2
            hole_trans = pya.Trans(Trans.R0, hole_x,hole_y)
            hole_t = hole_poly.transformed(hole_trans)
            hole.insert(hole_t) 
        
    phc = Si_slab - hole
    phc = phc + ruler
    self.cell.shapes(LayerSiN).insert(phc)
"""


class Hole_cell_half(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(Hole_cell_half, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Cavity Hole Cell_a%s-r%.3f" % (self.a, self.r)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        # Fetch all the parameters:
        a = self.a / dbu
        r = self.r / dbu

        # function to generate points to create a circle
        def hexagon_hole_half(a, r):
            npts = 10
            theta_div = math.pi / 3
            theta_div_hole = math.pi / npts
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            for i in range(0, npts + 1):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            r * math.cos(math.pi / 2 - i * theta_div_hole),
                            r * math.sin(math.pi / 2 - i * theta_div_hole),
                        )
                    )
                )
            return pts

        hole_cell = pya.Region()
        hole_cell_pts = hexagon_hole_half(a, r)
        hole_cell_poly_half = pya.Polygon(hole_cell_pts)

        # hole_cell.insert(hole_cell_poly_0)

        self.cell.shapes(LayerSiN).insert(hole_cell_poly_half)


class Hexagon_cell_half(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(Hexagon_cell_half, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Cavity Hole Cell_a%s-r%.3f" % (self.a, self.r)

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        # Fetch all the parameters:
        a = self.a / dbu
        r = self.r / dbu

        # function to generate points to create a circle
        def hexagon_half(a):
            theta_div = math.pi / 3
            triangle_length = a / math.sqrt(3)
            pts = []
            for i in range(0, 4):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            triangle_length * math.cos(i * theta_div - math.pi / 2),
                            triangle_length * math.sin(i * theta_div - math.pi / 2),
                        )
                    )
                )
            return pts

        hexagon_pts = hexagon_half(a)
        hexagon_cell_poly_half = pya.Polygon(hexagon_pts)

        # hole_cell.insert(hole_cell_poly_0)

        self.cell.shapes(LayerSiN).insert(hexagon_cell_poly_half)


class wg_triangle_tapers(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the strip waveguide taper.
    """

    def __init__(self):
        # Important: initialize the super class
        super(wg_triangle_tapers, self).__init__()

        # declare the parameters
        self.param(
            "tri_base", self.TypeDouble, "Triangle Base (microns)", default=0.363
        )
        self.param(
            "tri_height", self.TypeDouble, "Triangle Height (microns)", default=0.426
        )
        self.param(
            "taper_wg_length", self.TypeDouble, "Waveguide Length (microns)", default=5
        )
        self.param("wg_width", self.TypeDouble, "Waveguide Width (microns)", default=1)
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("silayer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "waveguide_triangular_tapers_%.3f-%.3f" % (
            self.taper_wg_length,
            self.wg_width,
        )

    def can_create_from_shape_impl(self):
        return False

    def produce(self, layout, layers, parameters, cell):
        """
        coerce parameters (make consistent)
        """
        self._layers = layers
        self.cell = cell
        self._param_values = parameters
        self.layout = layout
        shapes = self.cell.shapes

        # cell: layout cell to place the layout
        # LayerSiN: which layer to use
        # w: waveguide width
        # length units in dbu

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.silayer
        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        base = int(round(self.tri_base / dbu))
        height = int(round(self.tri_height / dbu))
        l = int(round(self.taper_wg_length / dbu))
        w = int(round(self.wg_width / dbu))

        pts = [
            Point(-l, w / 2),
            Point(-base, w / 2),
            Point(0, w / 2 + height),
            Point(0, -(w / 2 + height)),
            Point(-base, -w / 2),
            Point(-l, -w / 2),
        ]
        shapes(LayerSiN).insert(Polygon(pts))

        # Pins on the bus waveguide side:
        pin_length = 200
        if l < pin_length + 1:
            pin_length = int(l / 3)
            pin_length = math.ceil(pin_length / 2.0) * 2
        if pin_length == 0:
            pin_length = 2

        t = Trans(Trans.R0, -l, 0)
        pin = pya.Path([Point(-pin_length / 2, 0), Point(pin_length / 2, 0)], w)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin1", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = Trans(Trans.R0, 0, 0)
        pin_t = pin.transformed(t)
        shapes(LayerPinRecN).insert(pin_t)
        text = Text("pin2", t)
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        # box1 = Box(w/2+height, -(w/2+height), -l, -1)
        # shapes(LayerDevRecN).insert(box1)

        return "wg_triangle_taper"


def layout_waveguide_abs(cell, layer, points, w, radius):
    # create a path, then convert to a polygon waveguide with bends
    # cell: cell into which to place the waveguide
    # layer: layer to draw on
    # points: array of vertices, absolute coordinates on the current cell
    # w: waveguide width

    # example usage:
    # cell = pya.Application.instance().main_window().current_view().active_cellview().cell
    # LayerSi = LayerInfo(1, 0)
    # points = [ [15, 2.75], [30, 2.75] ]  # units of microns.
    # layout_waveguide_abs(cell, LayerSi, points, 0.5, 10)

    if MODULE_NUMPY:
        # numpy version
        points = n.array(points)
        start_point = points[0]
        points = points - start_point
    else:
        # without numpy:
        start_point = []
        start_point.append(points[0][0])
        start_point.append(points[0][1])
        for i in range(0, 2):
            for j in range(0, len(points)):
                points[j][i] -= start_point[i]

    layout_waveguide_rel(cell, layer, start_point, points, w, radius)


def layout_waveguide_rel(cell, layer, start_point, points, w, radius):
    # create a path, then convert to a polygon waveguide with bends
    # cell: cell into which to place the waveguide
    # layer: layer to draw on
    # start_point: starting vertex for the waveguide
    # points: array of vertices, relative to start_point
    # w: waveguide width

    # example usage:
    # cell = pya.Application.instance().main_window().current_view().active_cellview().cell
    # LayerSi = LayerInfo(1, 0)
    # points = [ [15, 2.75], [30, 2.75] ]  # units of microns.
    # layout_waveguide_rel(cell, LayerSi, [0,0], points, 0.5, 10)

    # print("* layout_waveguide_rel(%s, %s, %s, %s)" % (cell.name, layer, w, radius) )

    ly = cell.layout()
    dbu = cell.layout().dbu

    start_point = [start_point[0] / dbu, start_point[1] / dbu]

    a1 = []
    for p in points:
        a1.append(pya.DPoint(float(p[0]), float(p[1])))

    wg_path = pya.DPath(a1, w)

    npoints = points_per_circle(radius / dbu)
    param = {
        "npoints": npoints,
        "radius": float(radius),
        "path": wg_path,
        "layer": layer,
    }

    pcell = ly.create_cell("ROUND_PATH", "Basic", param)

    # Configure the cell location
    trans = Trans(Point(start_point[0], start_point[1]))

    # Place the PCell
    cell.insert(pya.CellInstArray(pcell.cell_index(), trans))


class H0c_Test_Structure(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the test structure with grating couplers and waveguides and a photonic crystal cavity
    """

    def __init__(self):
        # Important: initialize the super class
        super(H0c_Test_Structure, self).__init__()

        # taper/wg parameters
        self.param(
            "tri_base", self.TypeDouble, "Taper Triangle Base (microns)", default=0.363
        )
        self.param(
            "tri_height",
            self.TypeDouble,
            "Taper Triangle Height (microns)",
            default=0.426,
        )
        self.param(
            "taper_wg_length", self.TypeDouble, "Taper Length (microns)", default=5
        )
        self.param(
            "wg_bend_radius",
            self.TypeDouble,
            "Waveguide Bend Radius (microns)",
            default=15,
        )

        # photonic crystal cavity
        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=30
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=3
        )
        self.param("S1x", self.TypeDouble, "S1x shift", default=0.28)
        self.param("S2x", self.TypeDouble, "S2x shift", default=0.193)
        self.param("S3x", self.TypeDouble, "S3x shift", default=0.194)
        self.param("S4x", self.TypeDouble, "S4x shift", default=0.162)
        self.param("S5x", self.TypeDouble, "S5x shift", default=0.113)
        self.param("S1y", self.TypeDouble, "S1y shift", default=-0.016)
        self.param("S2y", self.TypeDouble, "S2y shift", default=0.134)
        self.param(
            "phc_xdis",
            self.TypeDouble,
            "Distance from GC to middle of Cavity",
            default=35,
        )

        # GC parameters
        self.param(
            "wavelength", self.TypeDouble, "Design Wavelength (micron)", default=2.9
        )
        self.param("n_t", self.TypeDouble, "Fiber Mode", default=1.0)
        self.param("n_e", self.TypeDouble, "Grating Index Parameter", default=3.1)
        self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default=20.0)
        self.param(
            "grating_length", self.TypeDouble, "Grating Length (micron)", default=32.0
        )
        self.param(
            "taper_length", self.TypeDouble, "Taper Length (micron)", default=32.0
        )
        self.param("dc", self.TypeDouble, "Duty Cycle", default=0.488193)
        self.param("period", self.TypeDouble, "Grating Period", default=1.18939)
        self.param("ff", self.TypeDouble, "Fill Factor", default=0.244319)
        self.param("t", self.TypeDouble, "Waveguide Width (micron)", default=1.0)
        self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default=8.0)

        # Layer Parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        a = self.a
        n = self.n
        wg_dis = self.wg_dis
        phc_xdis = self.phc_xdis
        wg_bend_radius = self.wg_bend_radius
        wg_width = self.t

        if (wg_dis) % 2 == 0:
            length_slab_x = n * a
        else:
            length_slab_x = (n - 1) * a

        half_slab_x = length_slab_x / 2

        param_phc = {
            "a": self.a,
            "n": self.n,
            "r": self.r,
            "wg_dis": self.wg_dis,
            "S1x": self.S1x,
            "S2x": self.S2x,
            "S3x": self.S3x,
            "S4x": self.S4x,
            "S5x": self.S5x,
            "S1y": self.S1y,
            "S2y": self.S2y,
            "layer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_phc = ly.create_cell("H0 cavity with waveguide", "SiQL_PCells", param_phc)
        t_phc = Trans(
            Trans.R0,
            phc_xdis / dbu,
            (127) / dbu - (math.sqrt(3) / 2 * a * (wg_dis + 1)) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_phc.cell_index(), t_phc))

        param_GC = {
            "wavelength": self.wavelength,
            "n_t": self.n_t,
            "n_e": self.n_e,
            "angle_e": self.angle_e,
            "grating_length": self.grating_length,
            "taper_length": self.taper_length,
            "dc": self.dc,
            "period": self.period,
            "ff": self.ff,
            "t": self.t,
            "theta_c": self.theta_c,
            "layer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_GC = ly.create_cell("SWG Fibre Coupler", "SiQL_PCells", param_GC)
        t_GC = Trans(Trans.R0, 0, 0)
        instance = cell.insert(
            pya.CellInstArray(
                pcell_GC.cell_index(), t_GC, Point(0, 127 / dbu), Point(0, 0), 3, 1
            )
        )

        param_taper = {
            "tri_base": self.tri_base,
            "tri_height": self.tri_height,
            "taper_wg_length": self.taper_wg_length,
            "silayer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_taper = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper1 = Trans(Trans.R0, (phc_xdis - half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper.cell_index(), t_taper1))

        pcell_taper2 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper2 = Trans(Trans.R180, (phc_xdis + half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper2.cell_index(), t_taper2))

        pcell_taper3 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper3 = Trans(
            Trans.R180,
            (phc_xdis + half_slab_x) / dbu,
            (127 - 2 * (wg_dis + 1) * math.sqrt(3) / 2 * a) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_taper3.cell_index(), t_taper3))

        # gc middle to in port
        points = [[0, 127], [phc_xdis - half_slab_x - self.taper_wg_length, 127]]
        layout_waveguide_abs(cell, LayerSi, points, wg_width, wg_bend_radius)

        # gc top to through port
        points2 = [
            [0, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 127],
            [(phc_xdis + half_slab_x + self.taper_wg_length), 127],
        ]
        layout_waveguide_abs(cell, LayerSi, points2, wg_width, wg_bend_radius)

        # gc bottom to coupled port
        points3 = [
            [0, 0],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 0],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius,
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length),
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
        ]
        layout_waveguide_abs(cell, LayerSi, points3, wg_width, wg_bend_radius)


class H0c_oxide_Test_Structure(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the test structure with grating couplers and waveguides and a photonic crystal cavity
    """

    def __init__(self):
        # Important: initialize the super class
        super(H0c_oxide_Test_Structure, self).__init__()

        # taper/wg parameters
        self.param(
            "tri_base", self.TypeDouble, "Taper Triangle Base (microns)", default=0.363
        )
        self.param(
            "tri_height",
            self.TypeDouble,
            "Taper Triangle Height (microns)",
            default=0.426,
        )
        self.param(
            "taper_wg_length", self.TypeDouble, "Taper Length (microns)", default=5
        )
        self.param(
            "wg_bend_radius",
            self.TypeDouble,
            "Waveguide Bend Radius (microns)",
            default=15,
        )

        # photonic crystal cavity
        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.690)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=30
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.125)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=3
        )
        self.param("n_bus", self.TypeInt, "Bus number, 1 or 2 ", default=2)
        self.param("n_vertices", self.TypeInt, "Vertices of a hole", default=20)
        self.param("S1x", self.TypeDouble, "S1x shift", default=0.28)
        self.param("S2x", self.TypeDouble, "S2x shift", default=0.193)
        self.param("S3x", self.TypeDouble, "S3x shift", default=0.194)
        self.param("S4x", self.TypeDouble, "S4x shift", default=0.162)
        self.param("S5x", self.TypeDouble, "S5x shift", default=0.113)
        self.param("S1y", self.TypeDouble, "S1y shift", default=-0.016)
        self.param("S2y", self.TypeDouble, "S2y shift", default=0.134)
        self.param(
            "phc_xdis",
            self.TypeDouble,
            "Distance from GC to middle of Cavity",
            default=35,
        )

        # GC parameters
        self.param(
            "wavelength", self.TypeDouble, "Design Wavelength (micron)", default=2.9
        )
        self.param("n_t", self.TypeDouble, "Fiber Mode", default=1.0)
        self.param("n_e", self.TypeDouble, "Grating Index Parameter", default=3.1)
        self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default=20.0)
        self.param(
            "grating_length", self.TypeDouble, "Grating Length (micron)", default=32.0
        )
        self.param(
            "taper_length", self.TypeDouble, "Taper Length (micron)", default=32.0
        )
        self.param("dc", self.TypeDouble, "Duty Cycle", default=0.488193)
        self.param("period", self.TypeDouble, "Grating Period", default=1.18939)
        self.param("ff", self.TypeDouble, "Fill Factor", default=0.244319)
        self.param("t", self.TypeDouble, "Waveguide Width (micron)", default=1.0)
        self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default=8.0)

        # Layer Parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)

        a = self.a
        n = self.n
        wg_dis = self.wg_dis
        phc_xdis = self.phc_xdis
        wg_bend_radius = self.wg_bend_radius
        wg_width = self.t

        if (wg_dis) % 2 == 0:
            length_slab_x = n * a
        else:
            length_slab_x = (n - 1) * a

        half_slab_x = length_slab_x / 2

        param_phc = {
            "a": self.a,
            "n": self.n,
            "r": self.r,
            "wg_dis": self.wg_dis,
            "layer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_phc = ly.create_cell(
            "H0 cavity with waveguide, no etching", "SiQL_PCells", param_phc
        )
        t_phc = Trans(
            Trans.R0,
            phc_xdis / dbu,
            (127) / dbu - (math.sqrt(3) / 2 * a * (wg_dis + 1)) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_phc.cell_index(), t_phc))

        param_GC = {
            "wavelength": self.wavelength,
            "n_t": self.n_t,
            "n_e": self.n_e,
            "angle_e": self.angle_e,
            "grating_length": self.grating_length,
            "taper_length": self.taper_length,
            "dc": self.dc,
            "period": self.period,
            "ff": self.ff,
            "t": self.t,
            "theta_c": self.theta_c,
            "layer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_GC = ly.create_cell("SWG Fibre Coupler", "SiQL_PCells", param_GC)
        t_GC = Trans(Trans.R0, 0, 0)
        instance = cell.insert(
            pya.CellInstArray(
                pcell_GC.cell_index(), t_GC, Point(0, 127 / dbu), Point(0, 0), 3, 1
            )
        )

        param_taper = {
            "tri_base": self.tri_base,
            "tri_height": self.tri_height,
            "taper_wg_length": self.taper_wg_length,
            "silayer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_taper = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper1 = Trans(Trans.R0, (phc_xdis - half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper.cell_index(), t_taper1))

        pcell_taper2 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper2 = Trans(Trans.R180, (phc_xdis + half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper2.cell_index(), t_taper2))

        pcell_taper3 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper3 = Trans(
            Trans.R180,
            (phc_xdis + half_slab_x) / dbu,
            (127 - 2 * (wg_dis + 1) * math.sqrt(3) / 2 * a) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_taper3.cell_index(), t_taper3))

        # gc middle to in port
        points = [[0, 127], [phc_xdis - half_slab_x - self.taper_wg_length, 127]]
        layout_waveguide_abs(cell, LayerSi, points, wg_width, wg_bend_radius)

        # gc top to through port
        points2 = [
            [0, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 127],
            [(phc_xdis + half_slab_x + self.taper_wg_length), 127],
        ]
        layout_waveguide_abs(cell, LayerSi, points2, wg_width, wg_bend_radius)

        # gc bottom to coupled port
        points3 = [
            [0, 0],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 0],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius,
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length),
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
        ]
        layout_waveguide_abs(cell, LayerSi, points3, wg_width, wg_bend_radius)


class L3c_Test_Structure(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the test structure with grating couplers and waveguides and a photonic crystal cavity
    """

    def __init__(self):
        # Important: initialize the super class
        super(L3c_Test_Structure, self).__init__()

        # taper parameters
        self.param(
            "tri_base", self.TypeDouble, "Taper Triangle Base (microns)", default=0.363
        )
        self.param(
            "tri_height",
            self.TypeDouble,
            "Taper Triangle Height (microns)",
            default=0.426,
        )
        self.param(
            "taper_wg_length", self.TypeDouble, "Taper Length (microns)", default=5
        )
        self.param("w", self.TypeDouble, "Waveguide Width", default=1.0)
        self.param(
            "wg_bend_radius",
            self.TypeDouble,
            "Waveguide Bend Radius (microns)",
            default=15,
        )

        # photonic crystal cavity
        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.720)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=34
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.181)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=3
        )
        self.param("S1x", self.TypeDouble, "S1x shift", default=0.337)
        self.param("S2x", self.TypeDouble, "S2x shift", default=0.27)
        self.param("S3x", self.TypeDouble, "S3x shift", default=0.088)
        self.param("S4x", self.TypeDouble, "S4x shift", default=0.323)
        self.param("S5x", self.TypeDouble, "S5x shift", default=0.0173)
        self.param(
            "phc_xdis",
            self.TypeDouble,
            "Distance from GC to middle of Cavity",
            default=35,
        )

        # GC parameters
        self.param(
            "wavelength", self.TypeDouble, "Design Wavelength (micron)", default=2.9
        )
        self.param("n_t", self.TypeDouble, "Fiber Mode", default=1.0)
        self.param("n_e", self.TypeDouble, "Grating Index Parameter", default=3.1)
        self.param("angle_e", self.TypeDouble, "Taper Angle (deg)", default=20.0)
        self.param(
            "grating_length", self.TypeDouble, "Grating Length (micron)", default=32.0
        )
        self.param(
            "taper_length", self.TypeDouble, "Taper Length (micron)", default=32.0
        )
        self.param("dc", self.TypeDouble, "Duty Cycle", default=0.488193)
        self.param("period", self.TypeDouble, "Grating Period", default=1.18939)
        self.param("ff", self.TypeDouble, "Fill Factor", default=0.244319)
        self.param("t", self.TypeDouble, "Waveguide Width (micron)", default=1.0)
        self.param("theta_c", self.TypeDouble, "Insertion Angle (deg)", default=8.0)

        # Layer Parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        a = self.a
        n = self.n
        wg_dis = self.wg_dis
        phc_xdis = self.phc_xdis
        wg_bend_radius = self.wg_bend_radius
        wg_width = self.w

        if wg_dis % 2 == 0:
            length_slab_x = (n - 1) * a
        else:
            length_slab_x = n * a

        half_slab_x = length_slab_x / 2

        param_phc = {
            "a": self.a,
            "n": self.n,
            "r": self.r,
            "wg_dis": self.wg_dis,
            "S1x": self.S1x,
            "S2x": self.S2x,
            "S3x": self.S3x,
            "S4x": self.S4x,
            "S5x": self.S5x,
            "layer": self.layer,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_phc = ly.create_cell("L3 cavity with waveguide", "SiQL_PCells", param_phc)
        t1 = Trans(
            Trans.R0,
            phc_xdis / dbu,
            (127) / dbu - (math.sqrt(3) / 2 * a * (wg_dis + 1)) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_phc.cell_index(), t1))

        param_GC = {
            "wavelength": self.wavelength,
            "n_t": self.n_t,
            "n_e": self.n_e,
            "angle_e": self.angle_e,
            "grating_length": self.grating_length,
            "taper_length": self.taper_length,
            "dc": self.dc,
            "period": self.period,
            "ff": self.ff,
            "t": self.t,
            "theta_c": self.theta_c,
            "layer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_GC = ly.create_cell("SWG Fibre Coupler", "SiQL_PCells", param_GC)
        t_GC = Trans(Trans.R0, 0, 0)
        instance = cell.insert(
            pya.CellInstArray(
                pcell_GC.cell_index(), t_GC, Point(0, 127 / dbu), Point(0, 0), 3, 1
            )
        )

        param_taper = {
            "tri_base": self.tri_base,
            "tri_height": self.tri_height,
            "taper_wg_length": self.taper_wg_length,
            "wg_width": self.w,
            "silayer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_taper = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper1 = Trans(Trans.R0, (phc_xdis - half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper.cell_index(), t_taper1))

        pcell_taper2 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper2 = Trans(Trans.R180, (phc_xdis + half_slab_x) / dbu, (127) / dbu)
        instance = cell.insert(pya.CellInstArray(pcell_taper2.cell_index(), t_taper2))

        pcell_taper3 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper3 = Trans(
            Trans.R180,
            (phc_xdis + half_slab_x) / dbu,
            (127 - 2 * (wg_dis + 1) * math.sqrt(3) / 2 * a) / dbu,
        )
        instance = cell.insert(pya.CellInstArray(pcell_taper3.cell_index(), t_taper3))

        # gc middle to in port
        points = [[0, 127], [phc_xdis - half_slab_x - self.taper_wg_length, 127]]
        layout_waveguide_abs(cell, LayerSi, points, wg_width, wg_bend_radius)

        # gc top to through port
        points2 = [
            [0, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 254],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 127],
            [(phc_xdis + half_slab_x + self.taper_wg_length), 127],
        ]
        layout_waveguide_abs(cell, LayerSi, points2, wg_width, wg_bend_radius)

        # gc bottom to coupled port
        points3 = [
            [0, 0],
            [(phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius, 0],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length) + wg_bend_radius,
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
            [
                (phc_xdis + half_slab_x + self.taper_wg_length),
                127 - 2 * (wg_dis + 1) * a * math.sqrt(3) / 2,
            ],
        ]
        layout_waveguide_abs(cell, LayerSi, points3, wg_width, wg_bend_radius)


class GC_to_GC_ref1(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the test structure with grating couplers and waveguides and a photonic crystal cavity
    """

    def __init__(self):
        # Important: initialize the super class
        super(GC_to_GC_ref1, self).__init__()

        # other waveguide parameters
        self.param(
            "wg_radius", self.TypeDouble, "Waveguide Radius (microns)", default=15
        )
        self.param(
            "wg_width", self.TypeDouble, "Waveguide x Distance (microns)", default=1
        )
        self.param(
            "wg_xdis", self.TypeDouble, "Waveguide x Distance (microns)", default=5
        )

        # Layer Parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        wg_r = self.wg_radius
        wg_w = self.wg_width
        wg_xdis = self.wg_xdis

        # uses the default parameters for the GC
        param_GC = {"layer": LayerSi, "pinrec": self.pinrec, "devrec": self.devrec}
        pcell_GC = ly.create_cell("SWG Fibre Coupler", "SiQL_PCells", param_GC)
        t_GC = Trans(Trans.R0, 0, 0)
        # instance = cell.insert(pya.place_cell(pcell_GC, t_GC, Point(0,127/dbu), Point(0,0), 2, 1))
        # instance=place_cell(cell,pcell_GC,[0.5,0.5])
        cell.insert(
            pya.CellInstArray(pcell_GC.cell_index(), pya.Trans(pya.Trans.R0, 0, 0))
        )
        print("test")
        return

        points = [[0, 0], [wg_r + wg_xdis, 0], [wg_r + wg_xdis, 127], [0, 127]]
        # layout_waveguide_abs(cell, LayerSi, points, wg_w, wg_r)


class PhC_W1wg_reference(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(PhC_W1wg_reference, self).__init__()

        # phc parameters
        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=30
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=2
        )
        self.param("n_vertices", self.TypeInt, "Vertices of a hole", default=32)
        self.param("etch_condition", self.TypeInt, "Etch = 1, No Etch = 2", default=1)
        self.param("phc_xdis", self.TypeDouble, "Distance to middle of phc", default=35)

        # other waveguide parameters
        self.param(
            "wg_radius", self.TypeDouble, "Waveguide Radius (microns)", default=15
        )
        self.param("wg_width", self.TypeDouble, "Waveguide Radius (microns)", default=1)

        # taper parameters
        self.param(
            "tri_base", self.TypeDouble, "Taper Triangle Base (microns)", default=0.363
        )
        self.param(
            "tri_height",
            self.TypeDouble,
            "Taper Triangle Height (microns)",
            default=0.426,
        )
        self.param(
            "taper_wg_length", self.TypeDouble, "Taper Length (microns)", default=5
        )

        # Layer Parameters
        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param(
            "etch", self.TypeLayer, "oxide etch layer", default=pya.LayerInfo(12, 0)
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)

        wg_r = self.wg_radius
        wg_w = self.wg_width
        phc_xdis = self.phc_xdis
        wg_dis = self.wg_dis
        a = self.a
        w = self.wg_width
        n = self.n
        etch_condition = self.etch_condition

        param_GC = {"layer": LayerSi, "pinrec": self.pinrec, "devrec": self.devrec}
        pcell_GC = ly.create_cell("SWG Fibre Coupler", "SiQL_PCells", param_GC)
        t_GC = Trans(Trans.R0, 0, 0)
        instance = cell.insert(
            pya.CellInstArray(
                pcell_GC.cell_index(), t_GC, Point(0, 127 / dbu), Point(0, 0), 2, 1
            )
        )

        if etch_condition == 1:
            param_phc = {
                "a": self.a,
                "n": self.n,
                "r": self.r,
                "n_bus": 1,
                "wg_dis": wg_dis,
                "etch_condition": etch_condition,
                "layer": LayerSi,
                "n_vertices": self.n_vertices,
                "pinrec": self.pinrec,
                "devrec": self.devrec,
                "etch": self.etch,
            }
            pcell_phc = ly.create_cell(
                "H0 cavity with waveguide", "SiQL_PCells", param_phc
            )
            t = Trans(
                Trans.R0,
                phc_xdis / dbu,
                0 - ((wg_dis + 1) * math.sqrt(3) / 2 * a) / dbu,
            )
            instance = cell.insert(pya.CellInstArray(pcell_phc.cell_index(), t))
        else:
            param_phc = {
                "a": self.a,
                "n": self.n,
                "r": self.r,
                "n_bus": 1,
                "wg_dis": wg_dis,
                "layer": LayerSi,
                "n_vertices": self.n_vertices,
                "pinrec": self.pinrec,
                "devrec": self.devrec,
                "etch": self.etch,
            }
            pcell_phc = ly.create_cell(
                "H0 cavity with waveguide, no etching", "SiQL_PCells", param_phc
            )
            t = Trans(
                Trans.R0,
                phc_xdis / dbu,
                0 - ((wg_dis + 1) * math.sqrt(3) / 2 * a) / dbu,
            )
            instance = cell.insert(pya.CellInstArray(pcell_phc.cell_index(), t))

        param_taper = {
            "tri_base": self.tri_base,
            "tri_height": self.tri_height,
            "taper_wg_length": self.taper_wg_length,
            "wg_width": w,
            "silayer": LayerSi,
            "pinrec": self.pinrec,
            "devrec": self.devrec,
        }
        pcell_taper = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper1 = Trans(Trans.R0, (phc_xdis - n * a / 2) / dbu, 0)
        instance = cell.insert(pya.CellInstArray(pcell_taper.cell_index(), t_taper1))

        pcell_taper2 = ly.create_cell(
            "Waveguide Triangle Tapers", "SiQL_PCells", param_taper
        )
        t_taper2 = Trans(Trans.R180, (phc_xdis + n * a / 2) / dbu, 0)
        instance = cell.insert(pya.CellInstArray(pcell_taper2.cell_index(), t_taper2))

        points = [[0, 0], [phc_xdis - n * a / 2 - self.taper_wg_length, 0]]
        layout_waveguide_abs(cell, LayerSi, points, wg_w, wg_r)

        points2 = [
            [phc_xdis + n * a / 2 + self.taper_wg_length, 0],
            [phc_xdis + n * a / 2 + self.taper_wg_length + wg_r, 0],
            [phc_xdis + n * a / 2 + self.taper_wg_length + wg_r, 127],
            [0, 127],
        ]
        layout_waveguide_abs(cell, LayerSi, points2, wg_w, wg_r)



from pya import *
import pya


class ebeam_dc_te1550(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the Directional Coupler
    by Lukas Chrostowski, 2018/09
    compact model in INTERCONNECT based on:
     - https://kx.lumerical.com/t/lcml-directional-coupler-based-on-lookup-table-lcml-dc-strip-1550-lookuptable/2094
     - only parameterized for Lc, for the 500 x 220 nm waveguide with 200 nm gap, and 5 micron radius
     could be improved:
     - https://kx.lumerical.com/t/lcml-directional-coupler-based-on-analytical-functions-lcml-dc-strip-1550-analytical/2091

    """

    def __init__(self):
        # Important: initialize the super class
        super(ebeam_dc_te1550, self).__init__()
        from SiEPIC.utils import get_technology_by_name

        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param("Lc", self.TypeDouble, "Coupler Length", default=10.0)
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "ebeam_dc_te1550(Lc=" + ("%.3f" % self.Lc) + ")"

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        # Fixed PCell parameters
        port_spacing = 2000  # spacing of the two ports, determines the angle required by the s-bend.
        r = 5000  # radius
        w = 500  # waveguide width
        g = 200  # gap

        from math import pi, cos, sin, acos
        from SiEPIC.utils import arc_wg, arc_wg_xy
        from SiEPIC._globals import PIN_LENGTH

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes
        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        TextLayerN = ly.layer(self.textl)

        Lc = int(round(self.Lc / dbu))

        # Create the parallel waveguides
        if Lc > 0:
            wg1 = Box(-Lc / 2, -w / 2 + (w + g) / 2, Lc / 2, w / 2 + (w + g) / 2)
            shapes(LayerSiN).insert(wg1)
            wg1 = Box(-Lc / 2, -w / 2 - (w + g) / 2, Lc / 2, w / 2 - (w + g) / 2)
            shapes(LayerSiN).insert(wg1)

        dc_angle = acos((r - abs(port_spacing / 2)) / r) * 180 / pi

        # bottom S-bends
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(Lc / 2, -r - (w + g) / 2, r, w, 90 - dc_angle, 90, dbu=dbu)
        )
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(-Lc / 2, -r - (w + g) / 2, r, w, 90, 90 + dc_angle, dbu=dbu)
        )
        y_bottom = round(-2 * (1 - cos(dc_angle / 180.0 * pi)) * r) - (w + g) / 2
        x_bottom = round(2 * sin(dc_angle / 180.0 * pi) * r)
        t = Trans(Trans.R0, -x_bottom - Lc / 2, y_bottom + r)
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, -90, -90 + dc_angle, dbu=dbu).transformed(t)
        )
        t = Trans(Trans.R0, x_bottom + Lc / 2, y_bottom + r)
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, -90 - dc_angle, -90, dbu=dbu).transformed(t)
        )

        # top S-bends
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(Lc / 2, r + (w + g) / 2, r, w, 270, 270 + dc_angle, dbu=dbu)
        )
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(-Lc / 2, r + (w + g) / 2, r, w, 270 - dc_angle, 270, dbu=dbu)
        )
        y_top = round(2 * (1 - cos(dc_angle / 180.0 * pi)) * r) + (w + g) / 2
        x_top = round(2 * sin(dc_angle / 180.0 * pi) * r)
        t = Trans(Trans.R0, -x_top - Lc / 2, y_top - r)
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, 90 - dc_angle, 90, dbu=dbu).transformed(t)
        )
        t = Trans(Trans.R0, x_top + Lc / 2, y_top - r)
        self.cell.shapes(LayerSiN).insert(
            arc_wg(r, w, 90, 90 + dc_angle, dbu=dbu).transformed(t)
        )

        # Pins on the bottom waveguide side:
        pin = Path(
            [
                Point(-x_bottom + PIN_LENGTH / 2 - Lc / 2, y_bottom),
                Point(-x_bottom - PIN_LENGTH / 2 - Lc / 2, y_bottom),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin1", Trans(Trans.R0, -x_bottom - Lc / 2, y_bottom))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(x_bottom - PIN_LENGTH / 2 + Lc / 2, y_bottom),
                Point(x_bottom + PIN_LENGTH / 2 + Lc / 2, y_bottom),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin3", Trans(Trans.R0, x_bottom + Lc / 2, y_bottom))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2

        # Pins on the top waveguide side:
        pin = Path(
            [
                Point(-x_bottom + PIN_LENGTH / 2 - Lc / 2, y_top),
                Point(-x_bottom - PIN_LENGTH / 2 - Lc / 2, y_top),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin2", Trans(Trans.R0, -x_bottom - Lc / 2, y_top))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        pin = Path(
            [
                Point(x_bottom - PIN_LENGTH / 2 + Lc / 2, y_top),
                Point(x_bottom + PIN_LENGTH / 2 + Lc / 2, y_top),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin)
        text = Text("pin4", Trans(Trans.R0, x_bottom + Lc / 2, y_top))
        shape = shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu
        shape.text_halign = 2

        # Merge all the waveguide shapes, to avoid any small gaps
        layer_temp = self.layout.layer(LayerInfo(913, 0))
        shapes_temp = self.cell.shapes(layer_temp)
        ShapeProcessor().merge(
            self.layout, self.cell, LayerSiN, shapes_temp, True, 0, True, True
        )
        self.cell.shapes(LayerSiN).clear()
        shapes_SiN = self.cell.shapes(LayerSiN)
        ShapeProcessor().merge(
            self.layout, self.cell, layer_temp, shapes_SiN, True, 0, True, True
        )
        self.cell.shapes(layer_temp).clear()

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        dev = Box(
            -x_bottom - Lc / 2,
            y_bottom - w / 2 - w,
            x_bottom + Lc / 2,
            y_top + w / 2 + w,
        )
        shapes(LayerDevRecN).insert(dev)

        # Compact model information
        t = Trans(Trans.R0, 0, -w)
        text = Text("Lumerical_INTERCONNECT_library=Design kits/ebeam", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r * 0.017
        t = Trans(Trans.R0, 0, 0)
        text = Text("Component=ebeam_dc_te1550", t)
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r * 0.017
        t = Trans(Trans.R0, 0, w)
        text = Text(
            "Spice_param:wg_width=%.3fu gap=%.3fu radius=%.3fu Lc=%.3fu"
            % (w * dbu, g * dbu, r * dbu, self.Lc),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = r * 0.017

        # print("Done drawing the layout for - ebeam_dc_te1550: %.3f" % ( self.Lc) )



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name
import math


class phc_H0c(pya.PCellDeclarationHelper):
    """
    Input: length, width
    """

    import numpy

    def __init__(self):
        # Important: initialize the super class
        super(phc_H0c, self).__init__()

        self.param("a", self.TypeDouble, "lattice constant (microns)", default=0.744)
        self.param(
            "n", self.TypeInt, "Number of holes in x and y direction", default=30
        )
        self.param("r", self.TypeDouble, "hole radius (microns)", default=0.179)
        self.param(
            "wg_dis", self.TypeInt, "Waveguide distance (number of holes)", default=3
        )
        self.param("n_bus", self.TypeInt, "Bus number, 1 or 2 ", default=2)
        self.param("n_vertices", self.TypeInt, "Vertices of a hole", default=32)
        self.param("S1x", self.TypeDouble, "S1x shift", default=0.28)
        self.param("S2x", self.TypeDouble, "S2x shift", default=0.193)
        self.param("S3x", self.TypeDouble, "S3x shift", default=0.194)
        self.param("S4x", self.TypeDouble, "S4x shift", default=0.162)
        self.param("S5x", self.TypeDouble, "S5x shift", default=0.113)
        self.param("S1y", self.TypeDouble, "S1y shift", default=-0.016)
        self.param("S2y", self.TypeDouble, "S2y shift", default=0.134)
        self.param(
            "etch_condition",
            self.TypeInt,
            "etch = 1 if etch box, etch = 2 if no etch box",
            default=1,
        )

        TECHNOLOGY = get_technology_by_name("EBeam")
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=TECHNOLOGY["Text"])
        self.param(
            "etch", self.TypeLayer, "oxide etch layer", default=pya.LayerInfo(12, 0)
        )

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "phc_H0c_a%s-r%.3f-wg_dis%.3f-n%.3f" % (
            self.a,
            self.r,
            self.wg_dis,
            self.n,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout

        LayerSi = self.layer
        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)
        LayerTextN = ly.layer(self.textl)
        LayerEtch = ly.layer(self.etch)

        # Fetch all the parameters:
        a = self.a / dbu
        r = self.r / dbu
        wg_dis = self.wg_dis + 1
        n_vertices = self.n_vertices
        n_bus = self.n_bus
        n = int(math.ceil(self.n / 2))
        Sx = [self.S1x, self.S2x, self.S3x, self.S4x, self.S5x]
        Sy = [self.S1y, 0, self.S2y]
        etch_condition = self.etch_condition

        if n_bus == 1:
            Sx = [0, 0, 0, 0, 0]
            Sy = [0, 0, 0]

        if wg_dis % 2 == 0:
            length_slab_x = (2 * n - 1) * a
        else:
            length_slab_x = 2 * n * a

        length_slab_y = 2 * (wg_dis + 10) * a * math.sqrt(3) / 2

        length_anchor_y = length_slab_y + 20 * a
        length_anchor_x = length_slab_x + 20 * a

        n_x = n
        n_y = wg_dis + 10

        # Define Si slab and hole region for future subtraction
        Si_slab = pya.Region()
        Si_slab.insert(
            pya.Box(
                -length_anchor_x / 2,
                -length_anchor_y / 2,
                length_anchor_x / 2,
                length_anchor_y / 2,
            )
        )
        hole = pya.Region()
        hole_r = r
        trench = pya.Region()

        # add the trenches for waveguide connection
        trench_width = 20 / dbu
        trench_height = 9 * a * math.sqrt(3) / 2
        wg_pos = a * math.sqrt(3) / 2 * wg_dis

        trench.insert(
            pya.Box(
                -trench_width - length_slab_x / 2,
                wg_pos - trench_height / 2,
                -length_slab_x / 2,
                wg_pos + trench_height / 2,
            )
        )
        trench.insert(
            pya.Box(
                length_slab_x / 2,
                wg_pos - trench_height / 2,
                trench_width + length_slab_x / 2,
                wg_pos + trench_height / 2,
            )
        )

        if n_bus == 2:
            wg_pos_2 = -a * math.sqrt(3) / 2 * wg_dis
            trench.insert(
                pya.Box(
                    length_slab_x / 2,
                    wg_pos_2 - trench_height / 2,
                    trench_width + length_slab_x / 2,
                    wg_pos_2 + trench_height / 2,
                )
            )

        # function to generate points to create a circle
        def circle(x, y, r):
            npts = n_vertices
            theta = 2 * math.pi / npts  # increment, in radians
            pts = []
            for i in range(0, npts):
                pts.append(
                    Point.from_dpoint(
                        pya.DPoint(
                            (x + r * math.cos(i * theta)) / 1,
                            (y + r * math.sin(i * theta)) / 1,
                        )
                    )
                )
            return pts

        # raster through all holes with shifts and waveguide

        hole_cell = circle(0, 0, hole_r)
        hole_poly = pya.Polygon(hole_cell)

        for j in range(-n_y, n_y + 1):
            if j % 2 == 0 and j != wg_dis:
                for i in range(-n_x, n_x + 1):
                    if j == -wg_dis and i > 3 and n_bus == 2:
                        None
                    elif j == 0 and i in (1, -1, 2, -2, 3, -3, 4, -4, 5, -5):
                        hole_x = abs(i) / i * (abs(i) - 0.5 + Sx[abs(i) - 1]) * a
                        hole_y = 0
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t = hole_poly.transformed(hole_trans)
                        hole.insert(hole_t)
                    elif i != 0:
                        hole_x = abs(i) / i * (abs(i) - 0.5) * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t = hole_poly.transformed(hole_trans)
                        hole.insert(hole_t)
            elif j % 2 == 1 and j != wg_dis:
                for i in range(-n_x, n_x + 1):
                    if j == -wg_dis and i > 3 and n_bus == 2:
                        None
                    elif i == 0 and j in (1, -1, 3, -3):
                        hole_x = 0
                        hole_y = (
                            j * a * (math.sqrt(3) / 2) + abs(j) / j * a * Sy[abs(j) - 1]
                        )
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t = hole_poly.transformed(hole_trans)
                        hole.insert(hole_t)
                    else:
                        hole_x = i * a
                        hole_y = j * a * math.sqrt(3) / 2
                        hole_trans = pya.Trans(Trans.R0, hole_x, hole_y)
                        hole_t = hole_poly.transformed(hole_trans)
                        hole.insert(hole_t)

        phc = Si_slab - hole - trench
        self.cell.shapes(LayerSiN).insert(phc)

        if etch_condition == 1:
            box_etch = pya.Box(
                -(length_slab_x / 2 - 3000),
                -(length_slab_y / 2 - 3000),
                length_slab_x / 2 - 3000,
                length_slab_y / 2 - 3000,
            )
            self.cell.shapes(LayerEtch).insert(box_etch)

        # Pins on the waveguide:
        pin_length = 200
        pin_w = a

        t = pya.Trans(Trans.R0, -length_slab_x / 2, wg_pos)
        pin = pya.Path(
            [pya.Point(pin_length / 2, 0), pya.Point(-pin_length / 2, 0)], pin_w
        )
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin1", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        t = pya.Trans(Trans.R0, length_slab_x / 2, wg_pos)
        pin = pya.Path(
            [pya.Point(-pin_length / 2, 0), pya.Point(pin_length / 2, 0)], pin_w
        )
        pin_t = pin.transformed(t)
        self.cell.shapes(LayerPinRecN).insert(pin_t)
        text = pya.Text("pin2", t)
        shape = self.cell.shapes(LayerPinRecN).insert(text)
        shape.text_size = 0.4 / dbu

        # pin for drop waveguide
        if n_bus == 2:
            t = pya.Trans(Trans.R0, length_slab_x / 2, -wg_pos)
            pin_t = pin.transformed(t)
            self.cell.shapes(LayerPinRecN).insert(pin_t)
            text = pya.Text("pin3", t)
            shape = self.cell.shapes(LayerPinRecN).insert(text)
            shape.text_size = 0.4 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        points = [[-length_slab_x / 2, 0], [length_slab_x / 2, 0]]
        points = [Point(each[0], each[1]) for each in points]
        path = Path(points, length_slab_y)
        self.cell.shapes(LayerDevRecN).insert(path.simple_polygon())



import pya
from pya import *
from SiEPIC.utils import get_technology_by_name


class Contra_DC_chirped(pya.PCellDeclarationHelper):
    """
    Author:   Mustafa Hammood
              Mustafa@siepic.com
    """

    def __init__(self):
        # Important: initialize the super class
        super(Contra_DC_chirped, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")

        # declare the parameters
        self.param(
            "number_of_periods", self.TypeInt, "Number of grating periods", default=3000
        )
        self.param(
            "grating_period",
            self.TypeDouble,
            "Start grating period (microns)",
            default=0.312,
        )
        self.param(
            "grating_period_end",
            self.TypeDouble,
            "Start grating period (microns)",
            default=0.328,
        )
        self.param("gap", self.TypeDouble, "Gap (microns)", default=0.15)
        self.param(
            "corrugation_width1",
            self.TypeDouble,
            "Waveguide 1 Corrugration width (microns)",
            default=0.03,
        )
        self.param(
            "corrugation_width2",
            self.TypeDouble,
            "Waveguide 2 Corrugration width (microns)",
            default=0.04,
        )
        self.param("AR", self.TypeBoolean, "Anti-Reflection Coating", default=True)
        self.param(
            "sinusoidal",
            self.TypeBoolean,
            "Grating Type (Rectangular=False, Sinusoidal=True)",
            default=False,
        )
        self.param("wg1_width", self.TypeDouble, "Waveguide 1 width", default=0.45)
        self.param("wg2_width", self.TypeDouble, "Waveguide 2 width", default=0.55)
        self.param("sbend", self.TypeBoolean, "Include S-bends", default=1)
        self.param("sbend_r", self.TypeDouble, "S-bend radius (microns)", default=15)
        self.param(
            "sbend_length", self.TypeDouble, "S-bend length (microns)", default=11
        )
        self.param("port_w", self.TypeDouble, "Port Waveguide width", default=0.5)
        self.param("index", self.TypeDouble, "Gaussian Index", default=0)
        self.param(
            "rib", self.TypeBoolean, "Use rib (ridge) waveguides?", default=False
        )
        self.param(
            "layer_rib",
            self.TypeLayer,
            "Rib Layer",
            default=TECHNOLOGY["Si - 90 nm rib"],
        )
        self.param("metal", self.TypeBoolean, "Use metal heaters?", default=False)
        self.param("heater_width", self.TypeDouble, "Metal heater width", default=3)
        self.param("metal_width", self.TypeDouble, "Metal contact width", default=5)
        self.param(
            "l_heater", self.TypeLayer, "Heater Layer", default=TECHNOLOGY["M1_heater"]
        )
        self.param(
            "l_metal",
            self.TypeLayer,
            "Metal contact Layer",
            default=TECHNOLOGY["M2_router"],
        )
        self.param("layer", self.TypeLayer, "Layer", default=TECHNOLOGY["Si"])
        self.param(
            "pinrecm",
            self.TypeLayer,
            "PinRecM Layer (metal)",
            default=TECHNOLOGY["PinRecM"],
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )

    #    self.param("textl", self.TypeLayer, "Text Layer", default = LayerInfo(10, 0))

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return "Contra_DC_chirped%.3f-%.3f" % (
            self.grating_period,
            self.grating_period_end,
        )

    def coerce_parameters_impl(self):
        pass

    def can_create_from_shape(self, layout, shape, layer):
        return False

    def produce_impl(self):
        import math

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSi = self.layer
        LayerSiN = ly.layer(LayerSi)
        LayerRib = ly.layer(self.layer_rib)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        from SiEPIC.extend import to_itype
        from SiEPIC.utils.layout import (
            layout_waveguide_sbend,
            layout_taper,
            make_pin,
            make_devrec_label,
        )

        # Draw the Bragg grating (bottom):
        # create chirped period array
        grating_period = int(round(self.grating_period / dbu))
        grating_period_end = int(round(self.grating_period_end / dbu))
        length = self.number_of_periods
        step = (grating_period - grating_period_end) * 1.0 / length

        grating_period = []
        box_width = []
        misalignment = []
        for i in range(length):
            grating_period.append(grating_period_end + i * step)
            grating_period[i] = round(grating_period_end + i * step)
            box_width.append(int(round(grating_period[i] / 2)))
            misalignment.append(int(round(grating_period[i] / 2)))

        shapes_wg = pya.Region()
        shapes_rib = pya.Region()
        w = to_itype(self.wg1_width, dbu)
        GaussianIndex = self.index
        half_w = w / 2
        half_corrugation_w = to_itype(self.corrugation_width1 / 2, dbu)

        N = self.number_of_periods
        if self.sinusoidal:
            x = 0
            npoints_sin = 40
            for i in range(0, self.number_of_periods):
                if i != 0:
                    x = x + grating_period[i]

                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = (
                    int(round(self.corrugation_width1 / 2 / dbu)) * profileFunction
                )
                box1 = Box(x, 0, x + box_width[i], half_w + profile)
                pts1 = [Point(x, 0)]
                pts3 = [Point(x + misalignment[i], 0)]
                for i1 in range(0, npoints_sin + 1):
                    x1 = i1 * 2 * math.pi / npoints_sin
                    y1 = round(profile * math.sin(x1))
                    x1 = round(x1 / 2 / math.pi * grating_period[i])
                    pts1.append(Point(x + x1, half_w + y1))
                    pts3.append(Point(x + misalignment[i] + x1, -half_w - y1))
                pts1.append(Point(x + grating_period[i], 0))
                pts3.append(Point(x + grating_period[i] + misalignment[i], 0))
                shapes(LayerSiN).insert(Polygon(pts1))
                shapes(LayerSiN).insert(Polygon(pts3))
            length = x + grating_period[i] + misalignment[i]
            # extra piece at the end:
            box2 = Box(x + grating_period[i], 0, length, half_w)
            shapes(LayerSiN).insert(box2)
            # extra piece at the beginning:
            box3 = Box(0, 0, misalignment[0], -half_w)
            shapes(LayerSiN).insert(box3)

        else:
            x = 0
            for i in range(0, self.number_of_periods):
                if i != 0:
                    x = x + grating_period[i]

                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = (
                    int(round(self.corrugation_width1 / 2 / dbu)) * profileFunction
                )
                box1 = Box(
                    x, 0, x + box_width[i], to_itype(half_w + profile, dbu * 1000)
                )
                box2 = Box(
                    x + box_width[i],
                    0,
                    x + grating_period[i],
                    to_itype(half_w - profile, dbu * 1000),
                )
                box3 = Box(
                    x + misalignment[i],
                    0,
                    x + box_width[i] + misalignment[i],
                    to_itype(-half_w - profile, dbu * 1000),
                )
                box4 = Box(
                    x + box_width[i] + misalignment[i],
                    0,
                    x + grating_period[i] + misalignment[i],
                    to_itype(-half_w + profile, dbu * 1000),
                )
                shapes(LayerSiN).insert(box1)
                shapes(LayerSiN).insert(box2)
                shapes(LayerSiN).insert(box3)
                shapes(LayerSiN).insert(box4)
            length = x + grating_period[i] + misalignment[i]
            # extra piece at the end:
            box2 = Box(x + grating_period[i], 0, length, half_w)
            shapes(LayerSiN).insert(box2)
            # extra piece at the beginning:
            box3 = Box(0, 0, misalignment[0], -half_w)
            shapes(LayerSiN).insert(box3)

        vertical_offset = (
            int(round(self.wg2_width / 2 / dbu))
            + int(round(self.gap / dbu))
            + int(round(self.wg1_width / 2 / dbu))
        )

        t = Trans(Trans.R0, 0, vertical_offset)

        # Draw the Bragg grating (top):
        w = to_itype(self.wg2_width, dbu)
        GaussianIndex = self.index
        half_w = w / 2
        half_corrugation_w = int(round(self.corrugation_width2 / 2 / dbu))

        N = self.number_of_periods
        if self.sinusoidal:
            npoints_sin = 40
            x = 0
            for i in range(0, self.number_of_periods):
                if i != 0:
                    x = x + grating_period[i]

                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = (
                    int(round(self.corrugation_width2 / 2 / dbu)) * profileFunction
                )
                box1 = Box(x, 0, x + box_width[i], -half_w + profile).transformed(t)
                pts1 = [Point(x, 0)]
                pts3 = [Point(x + misalignment[i], 0)]
                for i1 in range(0, npoints_sin + 1):
                    x1 = i1 * 2 * math.pi / npoints_sin
                    y1 = round(profile * math.sin(x1))
                    x1 = round(x1 / 2 / math.pi * grating_period[i])
                    #          print("x: %s, y: %s" % (x1,y1))
                    pts1.append(Point(x + x1, -half_w - y1))
                    pts3.append(Point(x + misalignment[i] + x1, +half_w + y1))
                pts1.append(Point(x + grating_period[i], 0))
                pts3.append(Point(x + grating_period[i] + misalignment[i], 0))
                shapes(LayerSiN).insert(Polygon(pts1).transformed(t))
                shapes(LayerSiN).insert(Polygon(pts3).transformed(t))
            length = x + grating_period[i] + misalignment[i]
            # extra piece at the end:
            box2 = Box(x + grating_period[i], 0, length, -half_w).transformed(t)
            shapes(LayerSiN).insert(box2)
            # extra piece at the beginning:
            box3 = Box(0, 0, misalignment[0], half_w).transformed(t)
            shapes(LayerSiN).insert(box3)

        else:
            x = 0
            for i in range(0, self.number_of_periods):
                if i != 0:
                    x = x + grating_period[i]

                profileFunction = math.exp(
                    -0.5 * (2 * GaussianIndex * (i - N / 2) / (N)) ** 2
                )
                profile = (
                    int(round(self.corrugation_width2 / 2 / dbu)) * profileFunction
                )
                box1 = Box(x, 0, x + box_width[i], -half_w - profile).transformed(t)
                box2 = Box(
                    x + box_width[i], 0, x + grating_period[i], -half_w + profile
                ).transformed(t)
                box3 = Box(
                    x + misalignment[i],
                    0,
                    x + box_width[i] + misalignment[i],
                    half_w + profile,
                ).transformed(t)
                box4 = Box(
                    x + box_width[i] + misalignment[i],
                    0,
                    x + grating_period[i] + misalignment[i],
                    half_w - profile,
                ).transformed(t)
                shapes(LayerSiN).insert(box1)
                shapes(LayerSiN).insert(box2)
                shapes(LayerSiN).insert(box3)
                shapes(LayerSiN).insert(box4)
            length = x + grating_period[i] + misalignment[i]
            box2 = Box(x + grating_period[i], 0, length, -half_w).transformed(t)
            shapes(LayerSiN).insert(box2)
            # extra piece at the beginning:
            box3 = Box(0, 0, misalignment[0], half_w).transformed(t)
            shapes(LayerSiN).insert(box3)

        # Create the pins on the waveguides, as short paths:
        w1 = to_itype(self.wg1_width, dbu)
        w2 = to_itype(self.wg2_width, dbu)
        y_offset_top = -w / 2 - to_itype(self.gap / 2, dbu)
        if self.sbend:
            port_w = to_itype(self.port_w, dbu)
            sbend_r = to_itype(self.sbend_r, dbu)
            sbend_length = to_itype(self.sbend_length, dbu)
            sbend_offset = (
                4 * port_w + port_w - (w1 + w2) / 2 - int(round(self.gap / dbu))
            )
            taper_length = 50 * max(abs(w1 - port_w), abs(w2 - port_w))

            t = Trans(Trans.R180, 0, 0)
            shapes_wg += layout_waveguide_sbend(
                self.cell,
                LayerSiN,
                t,
                w1,
                sbend_r,
                sbend_offset,
                sbend_length,
                insert=False,
            )
            t = Trans(Trans.R0, -sbend_length - taper_length, -sbend_offset)
            shapes_wg += layout_taper(
                self.cell, LayerSiN, t, port_w, w1, taper_length, insert=False
            )
            x = -sbend_length - taper_length
            y = -sbend_offset
            make_pin(self.cell, "opt1", [x, y], port_w, LayerPinRecN, 180)

            x = -taper_length - sbend_length
            y = vertical_offset
            make_pin(self.cell, "opt2", [x, y], port_w, LayerPinRecN, 180)

            x = length + sbend_length + taper_length
            y = -sbend_offset
            # i don't know why int(x) works but x doesnt
            # future debugger, email mustafa hammood :)
            make_pin(self.cell, "opt3", [int(x), y], port_w, LayerPinRecN, 0)

            x = length + taper_length + sbend_length
            y = vertical_offset
            make_pin(self.cell, "opt4", [int(x), y], port_w, LayerPinRecN, 0)

            t = Trans(Trans.R180, 0, vertical_offset)
            shapes_wg += layout_taper(
                self.cell, LayerSiN, t, w2, w2, sbend_length / 2, insert=False
            )
            t = Trans(Trans.R180, -sbend_length / 2, vertical_offset)
            shapes_wg += layout_taper(
                self.cell,
                LayerSiN,
                t,
                w2,
                port_w,
                taper_length + sbend_length / 2,
                insert=False,
            )

            t = Trans(Trans.R0, length, 0)
            shapes_wg += layout_waveguide_sbend(
                self.cell,
                LayerSiN,
                t,
                w1,
                sbend_r,
                -sbend_offset,
                sbend_length,
                insert=False,
            )
            t = Trans(Trans.R0, length + sbend_length, -sbend_offset)
            shapes_wg += layout_taper(
                self.cell, LayerSiN, t, w1, port_w, taper_length, insert=False
            )

            t = Trans(Trans.R0, length, vertical_offset)
            shapes_wg += layout_taper(
                self.cell, LayerSiN, t, w2, w2, sbend_length / 2, insert=False
            )
            t = Trans(Trans.R0, length + sbend_length / 2, vertical_offset)
            shapes_wg += layout_taper(
                self.cell,
                LayerSiN,
                t,
                w2,
                port_w,
                taper_length + sbend_length / 2,
                insert=False,
            )

        else:
            x = 0
            y = 0
            make_pin(self.cell, "opt1", [x, y], w1, LayerPinRecN, 180)

            y = vertical_offset
            make_pin(self.cell, "opt2", [x, y], w2, LayerPinRecN, 180)

            x = length
            y = 0
            make_pin(self.cell, "opt3", [int(x), y], w1, LayerPinRecN, 0)

            x = length
            y = vertical_offset
            make_pin(self.cell, "opt4", [int(x), y], w2, LayerPinRecN, 0)

        # Compact model information. Create the devrec library labels
        make_devrec_label(
            self.cell,
            "EBeam",
            "contra_directional_coupler_chirped",
            ly.layer(self.devrec),
        )

        text = Text(
            "Spice_param:number_of_periods=%s grating_period=%.3fu grating_period_end=%.3fu wg1_width=%.3fu wg2_width=%.3fu corrugation_width2=%.3fu corrugation_width2=%.3fu gap=%.3fu index=%.3f AR=%s sinusoidal=%s"
            % (
                self.number_of_periods,
                self.grating_period,
                self.grating_period_end,
                self.wg1_width,
                self.wg2_width,
                self.corrugation_width1,
                self.corrugation_width2,
                self.gap,
                self.index,
                int(self.AR),
                int(self.sinusoidal),
            ),
            t,
        )
        shape = shapes(LayerDevRecN).insert(text)
        shape.text_size = 0.1 / dbu

        if self.metal:
            # add heater element
            heater_width = to_itype(self.heater_width, dbu)
            box = pya.Box(
                pya.Point(0, heater_width / 2), pya.Point(length, -heater_width / 2)
            )
            shapes(ly.layer(self.l_heater)).insert(box)

            # add metal contact element (left and right)
            metal_width = to_itype(self.metal_width, dbu)
            box = pya.Box(
                pya.Point(0, heater_width / 2),
                pya.Point(metal_width, -heater_width / 2),
            )
            shapes(ly.layer(self.l_metal)).insert(box)
            box = pya.Box(
                pya.Point(length, heater_width / 2),
                pya.Point(length - metal_width, -heater_width / 2),
            )
            shapes(ly.layer(self.l_metal)).insert(box)

            # add metal contact pins
            x = metal_width / 2 * dbu
            y = -heater_width / 2 * dbu
            make_pin(
                self.cell, "elec1", [x, y], metal_width, ly.layer(self.pinrecm), 270
            )
            x = (length - metal_width / 2) * dbu
            make_pin(
                self.cell, "elec2", [x, y], metal_width, ly.layer(self.pinrecm), 270
            )

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        if self.sbend:
            box = pya.Box(
                pya.Point(-taper_length - sbend_length, vertical_offset + 3 * port_w),
                pya.Point(
                    length + taper_length + sbend_length,
                    y_offset_top - sbend_offset - 3 * port_w,
                ),
            )
        else:
            box = pya.Box(
                pya.Point(0, vertical_offset + 3 * (w1 + w2) / 2),
                pya.Point(length, y_offset_top - 3 * (w1 + w2) / 2),
            )
        DevRecBox = box
        shapes(LayerDevRecN).insert(box)

        # Draw the waveguide layer
        if self.rib == False:
            shapes(LayerSiN).insert(shapes_wg)
        else:  # turn shape into a rib waveguide
            region_devrec = Region(DevRecBox)
            region_devrec2 = Region(DevRecBox).size(1500)
            shapes_rib += shapes_wg
            shapeRib = (
                shapes_rib.size(1400) - shapes_wg - (region_devrec2 - region_devrec)
            )

            shapes(LayerRib).insert(shapeRib)
            shapes(LayerSiN).insert(shapes_wg)



"""
Debug sliver issue foundy by Susan at CMC

by Lukas Chrostowski 2024

Caused when a bezier curve ending is the end of the waveguide.
Caused by translate_from_normal(wg_pts, width/2 + (offset if turn > 0 else - offset))
Caused by Bezier curve, which returns:
  433, 12745
  220, 12748
  1,   12750
  0,   12750
 (length 111)
"""

designer_name = "Example"
top_cell_name = "EBeam_%s_MZI" % designer_name

import pya
from pya import *

import siepic_ebeam_pdk
from SiEPIC.scripts import connect_pins_with_waveguide, export_layout
from SiEPIC.utils.layout import new_layout

import os


def test_sliver():
    tech_name = "EBeam"
    from SiEPIC.utils import get_technology_by_name

    TECHNOLOGY = get_technology_by_name(tech_name)

    cell, ly = new_layout(tech_name, top_cell_name, GUI=True, overwrite=True)

    waveguide_type = "Strip TE 1550 nm, w=500 nm"

    cell_ebeam_y = ly.create_cell("ebeam_y_1550", tech_name)
    t = pya.Trans.from_s("r0 %s, %s" % (-7400, 0))
    instY1 = cell.insert(pya.CellInstArray(cell_ebeam_y.cell_index(), t))
    t = pya.Trans.from_s("r0 %s, %s" % (-7400, 10e3))
    instY2 = cell.insert(pya.CellInstArray(cell_ebeam_y.cell_index(), t))
    wg1 = connect_pins_with_waveguide(
        instY1, "opt2", instY2, "opt2", waveguide_type=waveguide_type
    )

    instY1.delete()
    instY2.delete()

    # check for sliver:
    wg1.cell.shapes(ly.layer(ly.TECHNOLOGY["Si"])).each
    polygon = [p for p in wg1.cell.shapes(ly.layer(ly.TECHNOLOGY["Si"])).each()][
        0
    ].polygon
    for p in [r for r in polygon.to_simple_polygon().each_point()]:
        if p.x < 0:
            raise Exception(" Polygon error: %s" % p)

    # Save
    path = os.path.dirname(os.path.realpath(__file__))
    filename = "test_waveguide_sliver"
    file_out = export_layout(cell, path, filename, relative_path="", format="oas")

    from SiEPIC.utils import klive

    klive.show(file_out, technology=tech_name, keep_position=True)


if __name__ == "__main__":
    test_sliver()



print('SiEPIC-EBeam-PDK v0.4.22 Python module: pymacros')

from . import (
    SiEPIC_EBeam_Library,
    SiEPIC_EBeam_Library_SiN,
    SiEPIC_EBeam_Library_Dream,
    SiEPIC_EBeam_Library_Beta,
    SiEPIC_EBeam_Library_ANT,
)
